/**
 * this source file released under the GNU Public Licence.
 * see the accompanying copyright.txt for more information
 * Copyright (C) 2000-2001 Ben Mazur
 */

/**
 * PgbCpu keeps the state of the CPU registers and executes
 * the current opcode.
 * 
 * It is a closed-box type of architecture for performance
 * purposes.  Sorry for all the private final methods...
 * 
 * The exec() method executes the current opcode and 
 * increases the program counter.  exec() also calls the 
 * memory cycle() function and processes any resulting 
 * interrupts.
 */
public final class PgbCpu {
	private final int				HALTON = -1;//0x0150;
	
	final static int				Z_FLAG = 0x80;       /* 1: Result is zero          */
	final static int				N_FLAG = 0x40;       /* 1: Subtraction occured     */
	final static int				H_FLAG = 0x20;       /* 1: Halfcarry/Halfborrow    */
	final static int				C_FLAG = 0x10;       /* 1: Carry/Borrow occured    */
	
	PgbMemory						mem;
	
	// registers
	private PgbRegisterB			AF;
	private PgbRegister				BC;
	private PgbRegister				DE;
	private PgbRegister				HL;
	private PgbRegisterW			SP;
	private PgbRegisterW			PC;
	
	private boolean					ime; // interrupt master enable
	
	// opcode lists
	private Opcode[]				opcodes;
	private Opcode[]				cbcodes;
	
	public PgbCpu(PgbMemory mem) {
		this.mem = mem;
		initOpcodes();
	}
	
	public void reset() {
		AF = new PgbRegisterB(0x0180);
		if(PgbSettings.system == PgbSettings.SYS_GBP) {
			AF = new PgbRegisterB(0xFF80);
		}
		if(PgbSettings.system == PgbSettings.SYS_GBC) {
			AF = new PgbRegisterB(0x1180);
		}
		BC = new PgbRegisterB(0x0013);
		DE = new PgbRegisterB(0x00D8);
		HL = new PgbRegisterW(0x014D);
		SP = new PgbRegisterW(0xFFFE);
		PC = new PgbRegisterW(0x0100);
		
		ime = false;
		
		PgbSettings.clockspeed = 4.194304;
	}
	
	public final void exec(int cte) {
		int cv;
		
		/*
		// debug halt
		if(PgbSettings.DEBUG) {
			if(PC.data == HALTON) {
				PgbSettings.paused = true;
				System.out.println("PC hit HALTON:" + Integer.toHexString(HALTON));
			}
		}
		*/
		
		
		/*
		if(PgbSettings.debuglevel == 5) {
			System.out.print("op:" + Integer.toHexString(opcode) + " pc:" + Integer.toHexString(PC.getR() - 1) + " af:" + Integer.toHexString(AF.getR()) + " bc:" + Integer.toHexString(BC.getR()) + " de:" + Integer.toHexString(DE.getR()) + " hl:" + Integer.toHexString(HL.getR()) + " sp:" + Integer.toHexString(SP.getR()) + " ");
			System.out.println("LCD y:" + Integer.toHexString(mem.video.ly) + " IF:" + Integer.toBinaryString(mem.IF));
		}
		*/
		
		while(cte > 0) {
			// process interrupts
			if(ime && (mem.IF & mem.IE) != 0) {
				interrupt(mem.IF & mem.IE);
			}
		
			// perform opcode
			cv = opcodes[(int)(mem.read(PC.data++) & 0xFF)].exec();
		
			// cycle all hardware and set the interrupt flag
			mem.cycle(cv);
			
			// loop
			cte -= cv;		
		}
	}
	
	private final void interrupt(int i) {
		if((i & mem.INT_VBLANK) == mem.INT_VBLANK) {
			// vBlank
			mem.IF &= ~mem.INT_VBLANK;
			di();
			rst(0x0040);
			return;
		}
		if((i & mem.INT_LCD) == mem.INT_LCD) {
			// LCD
			mem.IF &= ~mem.INT_LCD;
			di();
			rst(0x0048);
			return;
		}
		if((i & mem.INT_TIMER) == mem.INT_TIMER) {
			// timer overflow
			mem.IF &= ~mem.INT_TIMER;
			di();
			rst(0x0050);
			return;
		}
		if((i & mem.INT_SERIAL) == mem.INT_SERIAL) {
			// serial
			mem.IF &= ~mem.INT_SERIAL;
			di();
			rst(0x0058);
			return;
		}
		if((i & mem.INT_JOYPAD) == mem.INT_JOYPAD) {
			// joypad
			mem.IF &= ~mem.INT_JOYPAD;
			di();
			rst(0x0060);
			return;
		}
	}

	private abstract class Opcode {
		public abstract int exec();
	}
	private final class Op0x00 extends Opcode { // NOP
		public final int exec() {
			return 4;
		}
	}
	private final class Op0x01 extends Opcode { // LD   BC,nnnn
		public final int exec() {
			BC.setR(readWord());
			return 20;
		}
	}
	private final class Op0x02 extends Opcode { // LD   (BC),A
		public final int exec() {
			mem.write(BC.getR(), AF.getH());
			return 8;
		}
	}
	private final class Op0x03 extends Opcode { // INC  BC
		public final int exec() {
			incR(BC);
			return 8;
		}
	}
	private final class Op0x04 extends Opcode { // INC  B
		public final int exec() {
			incH(BC);
			return 4;
		}
	}
	private final class Op0x05 extends Opcode { // DEC  B
		public final int exec() {
			decH(BC);
			return 4;
		}
	}
	private final class Op0x06 extends Opcode { // LD   B,nn
		public final int exec() {
			BC.setH(readByte());
			return 8;
		}
	}
	private final class Op0x07 extends Opcode { // RLCA
		public final int exec() {
			rlcH(AF);
			setZ(false);
			return 4;
		}
	}
	private final class Op0x08 extends Opcode { // LD   (nnnn),SP
		public final int exec() {
			mem.writeWord(readWord(), SP.getR());
			return 20;
		}
	}
	private final class Op0x09 extends Opcode { // ADD  HL,BC
		public final int exec() {
			addR(HL, BC);
			return 12;
		}
	}
	private final class Op0x0A extends Opcode { // LD   A,(BC)
		public final int exec() {
			AF.setH(BC.getI(mem));
			return 8;
		}
	}
	private final class Op0x0B extends Opcode { // DEC  BC
		public final int exec() {
			decR(BC);
			return 8;
		}
	}
	private final class Op0x0C extends Opcode { // INC  C
		public final int exec() {
			incL(BC);
			return 4;
		}
	}
	private final class Op0x0D extends Opcode { // DEC  C
		public final int exec() {
			decL(BC);
			return 4;
		}
	}
	private final class Op0x0E extends Opcode { // LD   C,nn
		public final int exec() {
			BC.setL(readByte());
			return 8;
		}
	}
	private final class Op0x0F extends Opcode { // RRCA
		public final int exec() {
			rrcH(AF);
			setZ(false);
			return 4;
		}
	}
	private final class Op0x10 extends Opcode { // STOP
		public final int exec() {
			readByte();
			if(PgbSettings.system == PgbSettings.SYS_GBC) {
				// change speed
				if((mem.gbcGetSpeed() & 0x01) == 0x01) {
					PgbSettings.clockspeed = 8.388;
					mem.gbcSetSpeed((byte)0x80);
				} else {
					PgbSettings.clockspeed = 4.194;
					mem.gbcSetSpeed((byte)0x00);
				}
				//System.out.println("speed change : " + PgbSettings.clockspeed);
			} else {
				unsupported(0x10);
			}
			return 4;
		}
	}
	private final class Op0x11 extends Opcode { // LD   DE,nnnn
		public final int exec() {
			DE.setR(readWord());
			return 12;
		}
	}
	private final class Op0x12 extends Opcode { // LD   (DE),A
		public final int exec() {
			DE.setI(AF.getH(), mem);
			return 8;
		}
	}
	private final class Op0x13 extends Opcode { // INC  DE
		public final int exec() {
			incR(DE);
			return 8;
		}
	}
	private final class Op0x14 extends Opcode { // INC  D
		public final int exec() {
			incH(DE);
			return 4;
		}
	}
	private final class Op0x15 extends Opcode { // DEC  D
		public final int exec() {
			decH(DE);
			return 4;
		}
	}
	private final class Op0x16 extends Opcode { // LD   D,nn
		public final int exec() {
			DE.setH(readByte());
			return 8;
		}
	}
	private final class Op0x17 extends Opcode { // RLA
		public final int exec() {
			rlH(AF);
			setZ(false);
			return 4;
		}
	}
	private final class Op0x18 extends Opcode { // JR   disp
		public final int exec() {
			jr(readByte());
			return 12;
		}
	}
	private final class Op0x19 extends Opcode { // ADD  HL,DE
		public final int exec() {
			addR(HL, DE);
			return 16;
		}
	}
	private final class Op0x1A extends Opcode { // LD   A,(DE)
		public final int exec() {
			AF.setH(DE.getI(mem));
			return 8;
		}
	}
	private final class Op0x1B extends Opcode { // DEC  DE
		public final int exec() {
			decR(DE);
			return 4;
		}
	}
	private final class Op0x1C extends Opcode { // INC  E
		public final int exec() {
			incL(DE);
			return 4;
		}
	}
	private final class Op0x1D extends Opcode { // DEC  E
		public final int exec() {
			decL(DE);
			return 4;
		}
	}
	private final class Op0x1E extends Opcode { // LD   E,nn
		public final int exec() {
			DE.setL(readByte());
			return 8;
		}
	}
	private final class Op0x1F extends Opcode { // RRA
		public final int exec() {
			rrH(AF);
			setZ(false);
			return 4;
		}
	}
	private final class Op0x20 extends Opcode { // JR   NZ,disp
		public final int exec() {
			if(!getZ()) {
				jr(readByte());
				return 12;
			} else {
				readByte(); // and NOP it!
				return 8;
			}
		}
	}
	private final class Op0x21 extends Opcode { // LD   HL,nnnn
		public final int exec() {
			HL.setR(readWord());
			return 12;
		}
	}
	private final class Op0x22 extends Opcode { // LDI  (HL),A 
		public final int exec() {
			HL.setI(AF.getH(), mem);
			incR(HL);
			return 16;
		}
	}
	private final class Op0x23 extends Opcode { // INC  HL
		public final int exec() {
			incR(HL);
			return 8;
		}
	}
	private final class Op0x24 extends Opcode { // INC  H
		public final int exec() {
			incH(HL);
			return 4;
		}
	}
	private final class Op0x25 extends Opcode { // DEC  H
		public final int exec() {
			decH(HL);
			return 4;
		}
	}
	private final class Op0x26 extends Opcode { // LD   H,nn
		public final int exec() {
			HL.setH(readByte());
			return 8;
		}
	}
	private final class Op0x27 extends Opcode { // DAA
		public final int exec() {
			daa();
			return 4;
		}
	}
	private final class Op0x28 extends Opcode { // JR   Z,disp
		public final int exec() {
			if(getZ()) {
				jr(readByte());
				return 12;
			} else {
				readByte(); // and NOP it!
				return 8;
			}
		}
	}
	private final class Op0x29 extends Opcode { // ADD  HL,HL
		public final int exec() {
			addR(HL, HL);
			return 12;
		}
	}
	private final class Op0x2A extends Opcode { // LDI  A,(HL)
		public final int exec() {
			AF.setH(HL.getI(mem));
			incR(HL);
			return 16;
		}
	}
	private final class Op0x2B extends Opcode { // DEC HL
		public final int exec() {
			decR(HL);
			return 8;
		}
	}
	private final class Op0x2C extends Opcode { // INC  L
		public final int exec() {
			incL(HL);
			return 4;
		}
	}
	private final class Op0x2D extends Opcode { // DEC  L
		public final int exec() {
			decL(HL);
			return 4;
		}
	}
	private final class Op0x2E extends Opcode { // LD   L,nn
		public final int exec() {
			HL.setL(readByte());
			return 8;
		}
	}
	private final class Op0x2F extends Opcode { // CPL
		public final int exec() {
			cpl();
			return 4;
		}
	}
	private final class Op0x30 extends Opcode { // JR   NC,disp
		public final int exec() {
			if(!getC()) {
				jr(readByte());
				return 12;
			} else {
				readByte(); // and NOP it!
				return 8;
			}
		}
	}
	private final class Op0x31 extends Opcode { // LD   SP,nnnn
		public final int exec() {
			SP.setR(readWord());
			return 12;
		}
	}
	private final class Op0x32 extends Opcode { // LDD  (HL),A
		public final int exec() {
			HL.setI(AF.getH(), mem);
			decR(HL);
			return 16;
		}
	}
	private final class Op0x33 extends Opcode { // INC  SP
		public final int exec() {
			incR(SP);
			return 8;
		}
	}
	private final class Op0x34 extends Opcode { // INC  (HL)
		public final int exec() {
			incI(HL);
			return 12;
		}
	}
	private final class Op0x35 extends Opcode { // DEC  (HL)
		public final int exec() {
			decI(HL);
			return 12;
		}
	}
	private final class Op0x36 extends Opcode { // LD   (HL),nn
		public final int exec() {
			HL.setI(readByte(), mem);
			return 12;
		}
	}
	private final class Op0x37 extends Opcode { // SCF
		public final int exec() {
			setC(true);
			return 4;
		}
	}
	private final class Op0x38 extends Opcode { // JR   C,disp
		public final int exec() {
			if(getC()) {
				jr(readByte());
				return 12;
			} else {
				readByte(); // and NOP it!
				return 8;
			}
		}
	}
	private final class Op0x39 extends Opcode { // ADD  HL,SP
		public final int exec() {
			addR(HL, SP);
			return 12;
		}
	}
	private final class Op0x3A extends Opcode { // LDD  A,(HL)
		public final int exec() {
			AF.setH(HL.getI(mem));
			decR(HL);
			return 16;
		}
	}
	private final class Op0x3B extends Opcode { // DEC  SP
		public final int exec() {
			decR(SP);
			return 8;
		}
	}
	private final class Op0x3C extends Opcode { // INC  A
		public final int exec() {
			incH(AF);
			return 4;
		}
	}
	private final class Op0x3D extends Opcode { // DEC  A
		public final int exec() {
			decH(AF);
			return 4;
		}
	}
	private final class Op0x3E extends Opcode { // LD   A,nn
		public final int exec() {
			AF.setH(readByte());
			return 8;
		}
	}
	private final class Op0x3F extends Opcode { // CCF
		public final int exec() {
			setC(!getC());
			return 4;
		}
	}
	private final class Op0x40 extends Opcode { // LD   B,B
		public final int exec() {
			BC.setH(BC.getH());
			return 4;
		}
	}
	private final class Op0x41 extends Opcode { // LD   B,C
		public final int exec() {
			BC.setH(BC.getL());
			return 4;
		}
	}
	private final class Op0x42 extends Opcode { // LD   B,D
		public final int exec() {
			BC.setH(DE.getH());
			return 4;
		}
	}
	private final class Op0x43 extends Opcode { // LD   B,E
		public final int exec() {
			BC.setH(DE.getL());
			return 4;
		}
	}
	private final class Op0x44 extends Opcode { // LD   B,H 
		public final int exec() {
			BC.setH(HL.getH());
			return 4;
		}
	}
	private final class Op0x45 extends Opcode { // LD   B,L 
		public final int exec() {
			BC.setH(HL.getL());
			return 4;
		}
	}
	private final class Op0x46 extends Opcode { // LD   B,(HL) 
		public final int exec() {
			BC.setH(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0x47 extends Opcode { // LD   B,A 
		public final int exec() {
			BC.setH(AF.getH());
			return 4;
		}
	}
	private final class Op0x48 extends Opcode { // LD   C,B 
		public final int exec() {
			BC.setL(BC.getH());
			return 4;
		}
	}
	private final class Op0x49 extends Opcode { // LD   C,C 
		public final int exec() {
			BC.setL(BC.getL());
			return 4;
		}
	}
	private final class Op0x4A extends Opcode { // LD   C,D 
		public final int exec() {
			BC.setL(DE.getH());
			return 4;
		}
	}
	private final class Op0x4B extends Opcode { // LD   C,E 
		public final int exec() {
			BC.setL(DE.getL());
			return 4;
		}
	}
	private final class Op0x4C extends Opcode { // LD   C,H
		public final int exec() {
			BC.setL(HL.getH());
			return 4;
		}
	}
	private final class Op0x4D extends Opcode { // LD   C,L
		public final int exec() {
			BC.setL(HL.getL());
			return 4;
		}
	}
	private final class Op0x4E extends Opcode { // LD   C,(HL) 
		public final int exec() {
			BC.setL(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0x4F extends Opcode { // LD   C,A 
		public final int exec() {
			BC.setL(AF.getH());
			return 4;
		}
	}
	private final class Op0x50 extends Opcode { // LD   D,B 
		public final int exec() {
			DE.setH(BC.getH());
			return 4;
		}
	}
	private final class Op0x51 extends Opcode { // LD   D,C 
		public final int exec() {
			DE.setH(BC.getL());
			return 4;
		}
	}
	private final class Op0x52 extends Opcode { // LD   D,D 
		public final int exec() {
			DE.setH(DE.getH());
			return 4;
		}
	}
	private final class Op0x53 extends Opcode { // LD   D,E 
		public final int exec() {
			DE.setH(DE.getL());
			return 4;
		}
	}
	private final class Op0x54 extends Opcode { // LD   D,H 
		public final int exec() {
			DE.setH(HL.getH());
			return 4;
		}
	}
	private final class Op0x55 extends Opcode { // LD   D,L 
		public final int exec() {
			DE.setH(HL.getL());
			return 4;
		}
	}
	private final class Op0x56 extends Opcode { // LD   D,(HL)
		public final int exec() {
			DE.setH(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0x57 extends Opcode { // LD   D,A 
		public final int exec() {
			DE.setH(AF.getH());
			return 4;
		}
	}
	private final class Op0x58 extends Opcode { // LD   E,B 
		public final int exec() {
			DE.setL(BC.getH());
			return 4;
		}
	}
	private final class Op0x59 extends Opcode { // LD   E,C 
		public final int exec() {
			DE.setL(BC.getL());
			return 4;
		}
	}
	private final class Op0x5A extends Opcode { // LD   E,D 
		public final int exec() {
			DE.setL(DE.getH());
			return 4;
		}
	}
	private final class Op0x5B extends Opcode { // LD   E,E 
		public final int exec() {
			DE.setL(DE.getL());
			return 4;
		}
	}
	private final class Op0x5C extends Opcode { // LD   E,H 
		public final int exec() {
			DE.setL(HL.getH());
			return 4;
		}
	}
	private final class Op0x5D extends Opcode { // LD   E,L 
		public final int exec() {
			DE.setL(HL.getL());
			return 4;
		}
	}
	private final class Op0x5E extends Opcode { // LD   E,(HL)
		public final int exec() {
			DE.setL(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0x5F extends Opcode { // LD   E,A
		public final int exec() {
			DE.setL(AF.getH());
			return 8;
		}
	}
	private final class Op0x60 extends Opcode { // LD   H,B
		public final int exec() {
			HL.setH(BC.getH());
			return 4;
		}
	}
	private final class Op0x61 extends Opcode { // LD   H,C
		public final int exec() {
			HL.setH(BC.getL());
			return 4;
		}
	}
	private final class Op0x62 extends Opcode { // LD   H,D
		public final int exec() {
			HL.setH(DE.getH());
			return 4;
		}
	}
	private final class Op0x63 extends Opcode { // LD   H,E
		public final int exec() {
			HL.setH(DE.getL());
			return 4;
		}
	}
	private final class Op0x64 extends Opcode { // LD   H,H
		public final int exec() {
			HL.setH(HL.getH());
			return 4;
		}
	}
	private final class Op0x65 extends Opcode { // LD   H,L
		public final int exec() {
			HL.setH(HL.getL());
			return 4;
		}
	}
	private final class Op0x66 extends Opcode { // LD   H,(HL)
		public final int exec() {
			HL.setH(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0x67 extends Opcode { // LD   H,A
		public final int exec() {
			HL.setH(AF.getH());
			return 4;
		}
	}
	private final class Op0x68 extends Opcode { // LD   L,B
		public final int exec() {
			HL.setL(BC.getH());
			return 4;
		}
	}
	private final class Op0x69 extends Opcode { // LD   L,C
		public final int exec() {
			HL.setL(BC.getL());
			return 4;
		}
	}
	private final class Op0x6A extends Opcode { // LD   L,D
		public final int exec() {
			HL.setL(DE.getH());
			return 4;
		}
	}
	private final class Op0x6B extends Opcode { // LD   L,E
		public final int exec() {
			HL.setL(DE.getL());
			return 4;
		}
	}
	private final class Op0x6C extends Opcode { // LD   L,H
		public final int exec() {
			HL.setL(HL.getH());
			return 4;
		}
	}
	private final class Op0x6D extends Opcode { // LD   L,L
		public final int exec() {
			HL.setL(HL.getL());
			return 4;
		}
	}
	private final class Op0x6E extends Opcode { // LD   L,(HL)
		public final int exec() {
			HL.setL(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0x6F extends Opcode { // LD   L,A
		public final int exec() {
			HL.setL(AF.getH());
			return 4;
		}
	}
	private final class Op0x70 extends Opcode { // LD   (HL),B
		public final int exec() {
			mem.write(HL.getR(), BC.getH());
			return 8;
		}
	}
	private final class Op0x71 extends Opcode { // LD   (HL),C
		public final int exec() {
			mem.write(HL.getR(), BC.getL());
			return 8;
		}
	}
	private final class Op0x72 extends Opcode { // LD   (HL),D
		public final int exec() {
			mem.write(HL.getR(), DE.getH());
			return 8;
		}
	}
	private final class Op0x73 extends Opcode { // LD   (HL),E
		public final int exec() {
			mem.write(HL.getR(), DE.getL());
			return 8;
		}
	}
	private final class Op0x74 extends Opcode { // LD   (HL),H
		public final int exec() {
			mem.write(HL.getR(), HL.getH());
			return 8;
		}
	}
	private final class Op0x75 extends Opcode { // LD   (HL),L
		public final int exec() {
			mem.write(HL.getR(), HL.getL());
			return 8;
		}
	}
	private final class Op0x76 extends Opcode { // HALT
		public final int exec() {
			if(ime) {
				//decR(PC);
				//System.out.println("cpu HALTed.");
				mem.recalcCyclesLeft();
				return mem.cyclesLeft;
			} else {
				//System.out.println("cpu not HALTed because IME=" + ime + " and IE=" + mem.IE);
				return 4;
			}
		}
	}
	private final class Op0x77 extends Opcode { // LD   (HL),A
		public final int exec() {
			mem.write(HL.getR(), AF.getH());
			return 8;
		}
	}
	private final class Op0x78 extends Opcode { // LD   A,B
		public final int exec() {
			AF.setH(BC.getH());
			return 4;
		}
	}
	private final class Op0x79 extends Opcode { // LD   A,C
		public final int exec() {
			AF.setH(BC.getL());
			return 4;
		}
	}
	private final class Op0x7A extends Opcode { // LD   A,D
		public final int exec() {
			AF.setH(DE.getH());
			return 4;
		}
	}
	private final class Op0x7B extends Opcode { // LD   A,E
		public final int exec() {
			AF.setH(DE.getL());
			return 4;
		}
	}
	private final class Op0x7C extends Opcode { // LD   A,H
		public final int exec() {
			AF.setH(HL.getH());
			return 4;
		}
	}
	private final class Op0x7D extends Opcode { // LD   A,L
		public final int exec() {
			AF.setH(HL.getL());
			return 4;
		}
	}
	private final class Op0x7E extends Opcode { // LD   A,(HL)
		public final int exec() {
			AF.setH(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0x7F extends Opcode { // LD   A,A
		public final int exec() {
			AF.setH(AF.getH());
			return 4;
		}
	}
	private final class Op0x80 extends Opcode { // ADD  A,B
		public final int exec() {
			add(BC.getH());
			return 4;
		}
	}
	private final class Op0x81 extends Opcode { // ADD  A,C
		public final int exec() {
			add(BC.getL());
			return 4;
		}
	}
	private final class Op0x82 extends Opcode { // ADD  A,D
		public final int exec() {
			add(DE.getH());
			return 4;
		}
	}
	private final class Op0x83 extends Opcode { // ADD  A,E
		public final int exec() {
			add(DE.getL());
			return 4;
		}
	}
	private final class Op0x84 extends Opcode { // ADD  A,H
		public final int exec() {
			add(HL.getH());
			return 4;
		}
	}
	private final class Op0x85 extends Opcode { // ADD  A,L
		public final int exec() {
			add(HL.getL());
			return 4;
		}
	}
	private final class Op0x86 extends Opcode { // ADD  A,(HL)
		public final int exec() {
			add(HL.getI(mem));
			return 4;
		}
	}
	private final class Op0x87 extends Opcode { // ADD  A,A
		public final int exec() {
			add(AF.getH());
			return 4;
		}
	}
	private final class Op0x88 extends Opcode { // ADC  A,B
		public final int exec() {
			adc(BC.getH());
			return 4;
		}
	}
	private final class Op0x89 extends Opcode { // ADC  A,C
		public final int exec() {
			adc(BC.getL());
			return 4;
		}
	}
	private final class Op0x8A extends Opcode { // ADC  A,D
		public final int exec() {
			adc(DE.getH());
			return 4;
		}
	}
	private final class Op0x8B extends Opcode { // ADC  A,E
		public final int exec() {
			adc(DE.getL());
			return 4;
		}
	}
	private final class Op0x8C extends Opcode { // ADC  A,H
		public final int exec() {
			adc(HL.getH());
			return 4;
		}
	}
	private final class Op0x8D extends Opcode { // ADC  A,L
		public final int exec() {
			adc(HL.getL());
			return 4;
		}
	}
	private final class Op0x8E extends Opcode { // ADC  A,(HL)
		public final int exec() {
			adc(HL.getI(mem));
			return 4;
		}
	}
	private final class Op0x8F extends Opcode { // ADC  A,A
		public final int exec() {
			adc(AF.getH());
			return 4;
		}
	}
	private final class Op0x90 extends Opcode { // SUB  B
		public final int exec() {
			sub(BC.getH());
			return 4;
		}
	}
	private final class Op0x91 extends Opcode { // SUB  C
		public final int exec() {
			sub(BC.getL());
			return 4;
		}
	}
	private final class Op0x92 extends Opcode { // SUB  D
		public final int exec() {
			sub(DE.getH());
			return 4;
		}
	}
	private final class Op0x93 extends Opcode { // SUB  E
		public final int exec() {
			sub(DE.getL());
			return 4;
		}
	}
	private final class Op0x94 extends Opcode { // SUB  H
		public final int exec() {
			sub(HL.getH());
			return 4;
		}
	}
	private final class Op0x95 extends Opcode { // SUB  L
		public final int exec() {
			sub(HL.getL());
			return 4;
		}
	}
	private final class Op0x96 extends Opcode { // SUB  (HL)
		public final int exec() {
			sub(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0x97 extends Opcode { // SUB  A
		public final int exec() {
			sub(AF.getH());
			return 4;
		}
	}
	private final class Op0x98 extends Opcode { // SBC  A,B
		public final int exec() {
			sbc(BC.getH());
			return 4;
		}
	}
	private final class Op0x99 extends Opcode { // SBC  A,C
		public final int exec() {
			sbc(BC.getL());
			return 4;
		}
	}
	private final class Op0x9A extends Opcode { // SBC  A,D
		public final int exec() {
			sbc(DE.getH());
			return 4;
		}
	}
	private final class Op0x9B extends Opcode { // SBC  A,E
		public final int exec() {
			sbc(DE.getL());
			return 4;
		}
	}
	private final class Op0x9C extends Opcode { // SBC  A,H
		public final int exec() {
			sbc(HL.getH());
			return 4;
		}
	}
	private final class Op0x9D extends Opcode { // SBC  A,L
		public final int exec() {
			sbc(HL.getL());
			return 4;
		}
	}
	private final class Op0x9E extends Opcode { // SBC  A,(HL)
		public final int exec() {
			sbc(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0x9F extends Opcode { // SBC  A,A
		public final int exec() {
			sbc(AF.getH());
			return 4;
		}
	}
	private final class Op0xA0 extends Opcode { // AND  B
		public final int exec() {
			and(BC.getH());
			return 4;
		}
	}
	private final class Op0xA1 extends Opcode { // AND  C
		public final int exec() {
			and(BC.getL());
			return 4;
		}
	}
	private final class Op0xA2 extends Opcode { // AND  D
		public final int exec() {
			and(DE.getH());
			return 4;
		}
	}
	private final class Op0xA3 extends Opcode { // AND  E
		public final int exec() {
			and(DE.getL());
			return 4;
		}
	}
	private final class Op0xA4 extends Opcode { // AND  H
		public final int exec() {
			and(HL.getH());
			return 4;
		}
	}
	private final class Op0xA5 extends Opcode { // AND  L
		public final int exec() {
			and(HL.getL());
			return 4;
		}
	}
	private final class Op0xA6 extends Opcode { // AND  (HL)
		public final int exec() {
			and(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0xA7 extends Opcode { // AND  A
		public final int exec() {
			and(AF.getH());
			return 4;
		}
	}
	private final class Op0xA8 extends Opcode { // XOR  B
		public final int exec() {
			xor(BC.getH());
			return 4;
		}
	}
	private final class Op0xA9 extends Opcode { // XOR  C
		public final int exec() {
			xor(BC.getL());
			return 4;
		}
	}
	private final class Op0xAA extends Opcode { // XOR  D
		public final int exec() {
			xor(DE.getH());
			return 4;
		}
	}
	private final class Op0xAB extends Opcode { // XOR  E
		public final int exec() {
			xor(DE.getL());
			return 4;
		}
	}
	private final class Op0xAC extends Opcode { // XOR  H
		public final int exec() {
			xor(HL.getH());
			return 4;
		}
	}
	private final class Op0xAD extends Opcode { // XOR  L
		public final int exec() {
			xor(HL.getL());
			return 4;
		}
	}
	private final class Op0xAE extends Opcode { // XOR  (HL)
		public final int exec() {
			xor(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0xAF extends Opcode { // XOR  A
		public final int exec() {
			xor(AF.getH());
			return 4;
		}
	}
	private final class Op0xB0 extends Opcode { // OR   B
		public final int exec() {
			or(BC.getH());
			return 4;
		}
	}
	private final class Op0xB1 extends Opcode { // OR   C
		public final int exec() {
			or(BC.getL());
			return 4;
		}
	}
	private final class Op0xB2 extends Opcode { // OR   D
		public final int exec() {
			or(DE.getH());
			return 4;
		}
	}
	private final class Op0xB3 extends Opcode { // OR   E
		public final int exec() {
			or(DE.getL());
			return 4;
		}
	}
	private final class Op0xB4 extends Opcode { // OR   H
		public final int exec() {
			or(HL.getH());
			return 4;
		}
	}
	private final class Op0xB5 extends Opcode { // OR   L
		public final int exec() {
			or(HL.getL());
			return 4;
		}
	}
	private final class Op0xB6 extends Opcode { // OR   (HL)
		public final int exec() {
			or(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0xB7 extends Opcode { // OR   A
		public final int exec() {
			or(AF.getH());
			return 4;
		}
	}
	private final class Op0xB8 extends Opcode { // CP   B
		public final int exec() {
			cp(BC.getH());
			return 4;
		}
	}
	private final class Op0xB9 extends Opcode { // CP   C
		public final int exec() {
			cp(BC.getL());
			return 4;
		}
	}
	private final class Op0xBA extends Opcode { // CP   D
		public final int exec() {
			cp(DE.getH());
			return 4;
		}
	}
	private final class Op0xBB extends Opcode { // CP   E
		public final int exec() {
			cp(DE.getL());
			return 4;
		}
	}
	private final class Op0xBC extends Opcode { // CP   H
		public final int exec() {
			cp(HL.getH());
			return 4;
		}
	}
	private final class Op0xBD extends Opcode { // CP   L
		public final int exec() {
			cp(HL.getL());
			return 4;
		}
	}
	private final class Op0xBE extends Opcode { // CP   (HL)
		public final int exec() {
			cp(HL.getI(mem));
			return 8;
		}
	}
	private final class Op0xBF extends Opcode { // CP   A
		public final int exec() {
			cp(AF.getH());
			return 4;
		}
	}
	private final class Op0xC0 extends Opcode { // RET  NZ
		public final int exec() {
			if(!getZ()) {
				ret();
				return 20;
			} else {
				return 8;
			}
		}
	}
	private final class Op0xC1 extends Opcode { // POP  BC
		public final int exec() {
			pop(BC);
			return 12;
		}
	}
	private final class Op0xC2 extends Opcode { // JP   NZ,nnnn
		public final int exec() {
			if(!getZ()) {
				jp(readWord());
				return 16;
			} else {
				readWord();
				return 12;
			}
		}
	}
	private final class Op0xC3 extends Opcode { // JP nnnn
		public final int exec() {
			jp(readWord());
			return 16;
		}
	}
	private final class Op0xC4 extends Opcode { // CALL NZ,nnnn
		public final int exec() {
			if(!getZ()) {
				call(readWord());
				return 24;
			} else {
				readWord();
				return 12;
			}
		}
	}
	private final class Op0xC5 extends Opcode { // PUSH BC
		public final int exec() {
			push(BC);
			return 16;
		}
	}
	private final class Op0xC6 extends Opcode { // ADD  A,nn
		public final int exec() {
			add(readByte());
			return 8;
		}
	}
	private final class Op0xC7 extends Opcode { // RST  00H
		public final int exec() {
			rst(0x0000);
			return 16;
		}
	}
	private final class Op0xC8 extends Opcode { // RET Z
		public final int exec() {
			if(getZ()) {
				ret();
				return 20;
			} else {
				return 8;
			}
		}
	}
	private final class Op0xC9 extends Opcode { // RET
		public final int exec() {
			ret();
			return 16;
		}
	}
	private final class Op0xCA extends Opcode { // JP   Z,nnnn
		public final int exec() {
			if(getZ()) {
				jp(readWord());
				return 16;
			} else {
				readWord();
				return 12;
			}
		}
	}
	private final class Op0xCB extends Opcode { // --special--
		public final int exec() {
			return cbcodes[readByte()].exec();
		}
	}
	private final class Op0xCC extends Opcode { // CALL Z,nnnn
		public final int exec() {
			if(getZ()) {
				call(readWord());
				return 24;
			} else {
				readWord();
				return 12;
			}
		}
	}
	private final class Op0xCD extends Opcode { // CALL nnnn
		public final int exec() {
				call(readWord());
			return 24;
		}
	}
	private final class Op0xCE extends Opcode { // ADC  A,nn
		public final int exec() {
			adc(readByte());
			return 8;
		}
	}
	private final class Op0xCF extends Opcode { // RST  8H
		public final int exec() {
			rst(0x0008);
			return 16;
		}
	}
	private final class Op0xD0 extends Opcode { // RET  NC
		public final int exec() {
			if(!getC()) {
				ret();
				return 20;
			} else {
				return 8;
			}
		}
	}
	private final class Op0xD1 extends Opcode { // POP  DE
		public final int exec() {
			pop(DE);
			return 10;}}
	private final class Op0xD2 extends Opcode { // JP   NC,nnnn
		public final int exec() {
			if(!getC()) {
				jp(readWord());
				return 20;
			} else {
				readWord();
				return 8;
			}
		}
	}
	private final class Op0xD3 extends Opcode { // -
		public final int exec() {
			unsupported(0xD3);
			return 4;
		}
	}
	private final class Op0xD4 extends Opcode { // CALL NC,nnnn
		public final int exec() {
			if(!getC()) {
				call(readWord());
				return 24;
			} else {
				readWord();
				return 12;
			}
		}
	}
	private final class Op0xD5 extends Opcode { // PUSH DE
		public final int exec() {
			push(DE);
			return 16;
		}
	}
	private final class Op0xD6 extends Opcode { // SUB  nn
		public final int exec() {
			sub(readByte());
			return 8;
		}
	}
	private final class Op0xD7 extends Opcode { // RST  10H
		public final int exec() {
			rst(0x0010);
			return 16;
		}
	}
	private final class Op0xD8 extends Opcode { // RET  C
		public final int exec() {
			if(getC()) {
				ret();
				return 20;
			} else {
				return 8;
			}
		}
	}
	private final class Op0xD9 extends Opcode { // RETI
		public final int exec() {
			ret();
			ei();
			return 16;
		}
	}
	private final class Op0xDA extends Opcode { // JP   C,nnnn
		public final int exec() {
			if(getC()) {
				jp(readWord());
				return 16;
			} else {
				readWord();
				return 12;
			}
		}
	}
	private final class Op0xDB extends Opcode { // -
		public final int exec() {
			unsupported(0xDB);
			return 4;
		}
	}
	private final class Op0xDC extends Opcode { // CALL C,nnnn
		public final int exec() {
			if(getC()) {
				call(readWord());
				return 24;
			} else {
				readWord();
				return 12;
			}
		}
	}
	private final class Op0xDD extends Opcode { // -
		public final int exec() {
			unsupported(0xDD);
			return 4;
		}
	}
	private final class Op0xDE extends Opcode { // SBC  A,nn
		public final int exec() {
			sbc(readByte());
			return 8;
		}
	}
	private final class Op0xDF extends Opcode { // RST  18H
		public final int exec() {
			rst(0x0018);
			return 16;
		}
	}
	private final class Op0xE0 extends Opcode { // LD   ($FF00+nn),A
		public final int exec() {
			mem.write(0xFF00 + readByte(), AF.getH());
			return 12;
		}
	}
	private final class Op0xE1 extends Opcode { // POP  HL
		public final int exec() {
			pop(HL);
			return 12;
		}
	}
	private final class Op0xE2 extends Opcode { // LD   ($FF00+C),A
		public final int exec() {
			mem.write(0xFF00 + BC.getL(), AF.getH());
			return 8;
		}
	}
	private final class Op0xE3 extends Opcode { // -
		public final int exec() {
			unsupported(0xE3);
			return 4;
		}
	}
	private final class Op0xE4 extends Opcode { // -
		public final int exec() {
			unsupported(0xE4);
			return 4;
		}
	}
	private final class Op0xE5 extends Opcode { // PUSH HL
		public final int exec() {
			push(HL);
			return 16;
		}
	}
	private final class Op0xE6 extends Opcode { // AND  nn
		public final int exec() {
			and(readByte());
			return 8;
		}
	}
	private final class Op0xE7 extends Opcode { // RST  20H
		public final int exec() {
			rst(0x0020);
			return 16;
		}
	}
	private final class Op0xE8 extends Opcode { // ADD  SP,dd
		public final int exec() {
			SP.setR(SP.getR() + (byte)readByte());
			setZ(false);
			return 16;
		}
	}
	private final class Op0xE9 extends Opcode { // JP   (HL)
		public final int exec() {
			jp(HL.getR());
			return 4;
		}
	}
	private final class Op0xEA extends Opcode { // LD   (nnnn),A     
		public final int exec() {
			mem.write(readWord(), AF.getH());
			return 16;
		}
	}
	private final class Op0xEB extends Opcode { // -
		public final int exec() {
			unsupported(0xEB);
			return 4;
		}
	}
	private final class Op0xEC extends Opcode { // -
		public final int exec() {
			unsupported(0xEC);
			return 4;
		}
	}
	private final class Op0xED extends Opcode { // -
		public final int exec() {
			unsupported(0xED);
			return 4;
		}
	}
	private final class Op0xEE extends Opcode { // XOR  nn
		public final int exec() {
			xor(readByte());
			return 8;
		}
	}
	private final class Op0xEF extends Opcode { // RST  28H
		public final int exec() {
			rst(0x0028);
			return 16;
		}
	}
	private final class Op0xF0 extends Opcode { // LD   A,($FF00+nn) 
		public final int exec() {
			AF.setH(mem.read(0xFF00 + readByte()));
			return 8;
		}
	}
	private final class Op0xF1 extends Opcode { // POP  AF
		public final int exec() {
			pop(AF);
			return 12;
		}
	}
	private final class Op0xF2 extends Opcode { // LD   A,(C) 
		public final int exec() {
			AF.setH(mem.read(0xFF00 + BC.getL()));
			return 8;
		}
	}
	private final class Op0xF3 extends Opcode { // DI
		public final int exec() {
			di();
			return 4;
		}
	}
	private final class Op0xF4 extends Opcode { // -
		public final int exec() {
			unsupported(0xF4);
			return 4;
		}
	}
	private final class Op0xF5 extends Opcode { // PUSH AF
		public final int exec() {
			push(AF);
			return 16;
		}
	}
	private final class Op0xF6 extends Opcode { // OR   nn
		public final int exec() {
			or(readByte());
			return 8;
		}
	}
	private final class Op0xF7 extends Opcode { // RST  30H
		public final int exec() {
			rst(0x0030);
			return 16;
		}
	}
	private final class Op0xF8 extends Opcode { // LD   HL,SP+dd
		public final int exec() {
			HL.setR(SP.getR() + (byte)readByte());
			return 12;
		}
	}
	private final class Op0xF9 extends Opcode { // LD   SP,HL
		public final int exec() {
			SP.setR(HL.getR());
			return 8;
		}
	}
	private final class Op0xFA extends Opcode { // LD   A,(nnnn)
		public final int exec() {
			AF.setH(mem.read(readWord()));
			return 16;
		}
	}
	private final class Op0xFB extends Opcode { // ??? EI ???
		public final int exec() {
			ei();
			return 4;
		}
	}
	private final class Op0xFC extends Opcode { // -
		public final int exec() {
			unsupported(0xFC);
			return 4;
		}
	}
	private final class Op0xFD extends Opcode { // EI
		public final int exec() {
			ei();
			return 4;
		}
	}
	private final class Op0xFE extends Opcode { // CP   nn
		public final int exec() {
			cp(readByte());
			return 8;
		}
	}
	private final class Op0xFF extends Opcode { // RST  38H
		public final int exec() {
			rst(0x0038);
			return 16;
		}
	}
	private final class OpCB0x00 extends Opcode { // RLC  B 
		public final int exec() {
			rlcH(BC);
			return 4; }}
	private final class OpCB0x01 extends Opcode { // RLC  C 
		public final int exec() {
			rlcL(BC);
			return 4; }}
	private final class OpCB0x02 extends Opcode { // RLC  D 
		public final int exec() {
			rlcH(DE);
			return 4; }}
	private final class OpCB0x03 extends Opcode { // RLC  E 
		public final int exec() {
			rlcL(DE);
			return 4; }}
	private final class OpCB0x04 extends Opcode { // RLC  H 
		public final int exec() {
			rlcH(HL);
			return 4; }}
	private final class OpCB0x05 extends Opcode { // RLC  L 
		public final int exec() {
			rlcL(HL);
			return 4; }}
	private final class OpCB0x06 extends Opcode { // RLC  (HL) 
		public final int exec() {
			rlcI(HL.getR());
			return 4; }}
	private final class OpCB0x07 extends Opcode { // RLC  A 
		public final int exec() {
			rlcH(AF);
			return 4; }}
	private final class OpCB0x08 extends Opcode { // RRC  B 
		public final int exec() {
			rrcH(BC);
			return 4; }}
	private final class OpCB0x09 extends Opcode { // RRC  C 
		public final int exec() {
			rrcL(BC);
			return 4; }}
	private final class OpCB0x0A extends Opcode { // RRC  D 
		public final int exec() {
			rrcH(DE);
			return 4; }}
	private final class OpCB0x0B extends Opcode { // RRC  E 
		public final int exec() {
			rrcL(DE);
			return 4; }}
	private final class OpCB0x0C extends Opcode { // RRC  H 
		public final int exec() {
			rrcH(HL);
			return 4; }}
	private final class OpCB0x0D extends Opcode { // RRC  L 
		public final int exec() {
			rrcL(HL);
			return 4; }}
	private final class OpCB0x0E extends Opcode { // RRC  (HL) 
		public final int exec() {
			rrcI(HL.getR());
			return 4; }}
	private final class OpCB0x0F extends Opcode { // RRC  A 
		public final int exec() {
			rrcH(AF);
			return 4; }}
	private final class OpCB0x10 extends Opcode { // RL B 
		public final int exec() {
			rlH(BC);
			return 4; }}
	private final class OpCB0x11 extends Opcode { // RL C 
		public final int exec() {
			rlL(BC);
			return 4; }}
	private final class OpCB0x12 extends Opcode { // RL D 
		public final int exec() {
			rlH(DE);
			return 4; }}
	private final class OpCB0x13 extends Opcode { // RL E 
		public final int exec() {
			rlL(DE);
			return 4; }}
	private final class OpCB0x14 extends Opcode { // RL H 
		public final int exec() {
			rlH(HL);
			return 4; }}
	private final class OpCB0x15 extends Opcode { // RL L 
		public final int exec() {
			rlL(HL);
			return 4; }}
	private final class OpCB0x16 extends Opcode { // RL (HL) 
		public final int exec() {
			rlI(HL.getR());
			return 4; }}
	private final class OpCB0x17 extends Opcode { // RL A 
		public final int exec() {
			rlH(AF);
			return 4; }}
	private final class OpCB0x18 extends Opcode { // RR B 
		public final int exec() {
			rrH(BC);
			return 4; }}
	private final class OpCB0x19 extends Opcode { // RR C 
		public final int exec() {
			rrL(BC);
			return 4; }}
	private final class OpCB0x1A extends Opcode { // RR D 
		public final int exec() {
			rrH(DE);
			return 4; }}
	private final class OpCB0x1B extends Opcode { // RR E 
		public final int exec() {
			rrL(DE);
			return 4; }}
	private final class OpCB0x1C extends Opcode { // RR H 
		public final int exec() {
			rrH(HL);
			return 4; }}
	private final class OpCB0x1D extends Opcode { // RR L 
		public final int exec() {
			rrL(HL);
			return 4; }}
	private final class OpCB0x1E extends Opcode { // RR (HL) 
		public final int exec() {
			rrI(HL.getR());
			return 4; }}
	private final class OpCB0x1F extends Opcode { // RR A 
		public final int exec() {
			rrH(AF);
			return 4; }}
	private final class OpCB0x20 extends Opcode { // SLA B 
		public final int exec() {
			slaH(BC);
			return 4; }}
	private final class OpCB0x21 extends Opcode { // SLA C 
		public final int exec() {
			slaL(BC);
			return 4; }}
	private final class OpCB0x22 extends Opcode { // SLA D 
		public final int exec() {
			slaH(DE);
			return 4; }}
	private final class OpCB0x23 extends Opcode { // SLA E 
		public final int exec() {
			slaL(DE);
			return 4; }}
	private final class OpCB0x24 extends Opcode { // SLA H 
		public final int exec() {
			slaH(HL);
			return 4; }}
	private final class OpCB0x25 extends Opcode { // SLA L 
		public final int exec() {
			slaL(HL);
			return 4; }}
	private final class OpCB0x26 extends Opcode { // SLA (HL) 
		public final int exec() {
			slaI(HL.getR());
			return 4; }}
	private final class OpCB0x27 extends Opcode { // SLA A 
		public final int exec() {
			slaH(AF);
			return 4; }}
	private final class OpCB0x28 extends Opcode { // SRA B 
		public final int exec() {
			sraH(BC);
			return 4; }}
	private final class OpCB0x29 extends Opcode { // SRA C 
		public final int exec() {
			sraL(BC);
			return 4; }}
	private final class OpCB0x2A extends Opcode { // SRA D 
		public final int exec() {
			sraH(DE);
			return 4; }}
	private final class OpCB0x2B extends Opcode { // SRA E 
		public final int exec() {
			sraL(DE);
			return 4; }}
	private final class OpCB0x2C extends Opcode { // SRA H 
		public final int exec() {
			sraH(HL);
			return 4; }}
	private final class OpCB0x2D extends Opcode { // SRA L 
		public final int exec() {
			sraL(HL);
			return 4; }}
	private final class OpCB0x2E extends Opcode { // SRA (HL) 
		public final int exec() {
			sraI(HL.getR());
			return 4; }}
	private final class OpCB0x2F extends Opcode { // SRA A 
		public final int exec() {
			sraH(AF);
			return 4; }}
	private final class OpCB0x30 extends Opcode { // SWAP B 
		public final int exec() {
			swapH(BC);
			return 4; }}
	private final class OpCB0x31 extends Opcode { // SWAP C 
		public final int exec() {
			swapL(BC);
			return 4; }}
	private final class OpCB0x32 extends Opcode { // SWAP D 
		public final int exec() {
			swapH(DE);
			return 4; }}
	private final class OpCB0x33 extends Opcode { // SWAP E 
		public final int exec() {
			swapL(DE);
			return 4; }}
	private final class OpCB0x34 extends Opcode { // SWAP H 
		public final int exec() {
			swapH(HL);
			return 4; }}
	private final class OpCB0x35 extends Opcode { // SWAP L 
		public final int exec() {
			swapL(HL);
			return 4; }}
	private final class OpCB0x36 extends Opcode { // SWAP (HL) 
		public final int exec() {
			swapI(HL.getR());
			return 4; }}
	private final class OpCB0x37 extends Opcode { // SWAP A 
		public final int exec() {
			swapH(AF);
			return 4; }}
	private final class OpCB0x38 extends Opcode { // SRL B 
		public final int exec() {
			srlH(BC);
			return 4; }}
	private final class OpCB0x39 extends Opcode { // SRL C 
		public final int exec() {
			srlL(BC);
			return 4; }}
	private final class OpCB0x3A extends Opcode { // SRL D 
		public final int exec() {
			srlH(DE);
			return 4; }}
	private final class OpCB0x3B extends Opcode { // SRL E 
		public final int exec() {
			srlL(DE);
			return 4; }}
	private final class OpCB0x3C extends Opcode { // SRL H 
		public final int exec() {
			srlH(HL);
			return 4; }}
	private final class OpCB0x3D extends Opcode { // SRL L 
		public final int exec() {
			srlL(HL);
			return 4; }}
	private final class OpCB0x3E extends Opcode { // SRL (HL) 
		public final int exec() {
			srlI(HL.getR());
			return 4; }}
	private final class OpCB0x3F extends Opcode { // SRL A 
		public final int exec() {
			srlH(AF);
			return 4; }}
	private final class OpCB0x40 extends Opcode { // BIT  0,B 
		public final int exec() {
			bit(0, BC.getH());
			return 4; }}
	private final class OpCB0x41 extends Opcode { // BIT  0,C 
		public final int exec() {
			bit(0, BC.getL());
			return 4; }}
	private final class OpCB0x42 extends Opcode { // BIT  0,D 
		public final int exec() {
			bit(0, DE.getH());
			return 4; }}
	private final class OpCB0x43 extends Opcode { // BIT  0,E 
		public final int exec() {
			bit(0, DE.getL());
			return 4; }}
	private final class OpCB0x44 extends Opcode { // BIT  0,H 
		public final int exec() {
			bit(0, HL.getH());
			return 4; }}
	private final class OpCB0x45 extends Opcode { // BIT  0,L 
		public final int exec() {
			bit(0, HL.getL());
			return 4; }}
	private final class OpCB0x46 extends Opcode { // BIT  0,(HL) 
		public final int exec() {
			bit(0, HL.getI(mem));
			return 4; }}
	private final class OpCB0x47 extends Opcode { // BIT  0,A 
		public final int exec() {
			bit(0, AF.getH());
			return 4; }}
	private final class OpCB0x48 extends Opcode { // BIT  1,B 
		public final int exec() {
			bit(1, BC.getH());
			return 4; }}
	private final class OpCB0x49 extends Opcode { // BIT  1,C 
		public final int exec() {
			bit(1, BC.getL());
			return 4; }}
	private final class OpCB0x4A extends Opcode { // BIT  1,D 
		public final int exec() {
			bit(1, DE.getH());
			return 4; }}
	private final class OpCB0x4B extends Opcode { // BIT  1,E 
		public final int exec() {
			bit(1, DE.getL());
			return 4; }}
	private final class OpCB0x4C extends Opcode { // BIT  1,H 
		public final int exec() {
			bit(1, HL.getH());
			return 4; }}
	private final class OpCB0x4D extends Opcode { // BIT  1,L 
		public final int exec() {
			bit(1, HL.getL());
			return 4; }}
	private final class OpCB0x4E extends Opcode { // BIT  1,(HL) 
		public final int exec() {
			bit(1, HL.getI(mem));
			return 4; }}
	private final class OpCB0x4F extends Opcode { // BIT  1,A 
		public final int exec() {
			bit(1, AF.getH());
			return 4; }}
	private final class OpCB0x50 extends Opcode { // BIT  2,B 
		public final int exec() {
			bit(2, BC.getH());
			return 4; }}
	private final class OpCB0x51 extends Opcode { // BIT  2,C 
		public final int exec() {
			bit(2, BC.getL());
			return 4; }}
	private final class OpCB0x52 extends Opcode { // BIT  2,D 
		public final int exec() {
			bit(2, DE.getH());
			return 4; }}
	private final class OpCB0x53 extends Opcode { // BIT  2,E 
		public final int exec() {
			bit(2, DE.getL());
			return 4; }}
	private final class OpCB0x54 extends Opcode { // BIT  2,H 
		public final int exec() {
			bit(2, HL.getH());
			return 4; }}
	private final class OpCB0x55 extends Opcode { // BIT  2,L 
		public final int exec() {
			bit(2, HL.getL());
			return 4; }}
	private final class OpCB0x56 extends Opcode { // BIT  2,(HL) 
		public final int exec() {
			bit(2, HL.getI(mem));
			return 4; }}
	private final class OpCB0x57 extends Opcode { // BIT  2,A 
		public final int exec() {
			bit(2, AF.getH());
			return 4; }}
	private final class OpCB0x58 extends Opcode { // BIT  3,B 
		public final int exec() {
			bit(3, BC.getH());
			return 4; }}
	private final class OpCB0x59 extends Opcode { // BIT  3,C 
		public final int exec() {
			bit(3, BC.getL());
			return 4; }}
	private final class OpCB0x5A extends Opcode { // BIT  3,D 
		public final int exec() {
			bit(3, DE.getH());
			return 4; }}
	private final class OpCB0x5B extends Opcode { // BIT  3,E 
		public final int exec() {
			bit(3, DE.getL());
			return 4; }}
	private final class OpCB0x5C extends Opcode { // BIT  3,H 
		public final int exec() {
			bit(3, HL.getH());
			return 4; }}
	private final class OpCB0x5D extends Opcode { // BIT  3,L 
		public final int exec() {
			bit(3, HL.getL());
			return 4; }}
	private final class OpCB0x5E extends Opcode { // BIT  3,(HL) 
		public final int exec() {
			bit(3, HL.getI(mem));
			return 4; }}
	private final class OpCB0x5F extends Opcode { // BIT  3,A 
		public final int exec() {
			bit(3, AF.getH());
			return 4; }}
	private final class OpCB0x60 extends Opcode { // BIT  4,B 
		public final int exec() {
			bit(4, BC.getH());
			return 4; }}
	private final class OpCB0x61 extends Opcode { // BIT  4,C 
		public final int exec() {
			bit(4, BC.getL());
			return 4; }}
	private final class OpCB0x62 extends Opcode { // BIT  4,D 
		public final int exec() {
			bit(4, DE.getH());
			return 4; }}
	private final class OpCB0x63 extends Opcode { // BIT  4,E 
		public final int exec() {
			bit(4, DE.getL());
			return 4; }}
	private final class OpCB0x64 extends Opcode { // BIT  4,H 
		public final int exec() {
			bit(4, HL.getH());
			return 4; }}
	private final class OpCB0x65 extends Opcode { // BIT  4,L 
		public final int exec() {
			bit(4, HL.getL());
			return 4; }}
	private final class OpCB0x66 extends Opcode { // BIT  4,(HL) 
		public final int exec() {
			bit(4, HL.getI(mem));
			return 4; }}
	private final class OpCB0x67 extends Opcode { // BIT  4,A 
		public final int exec() {
			bit(4, AF.getH());
			return 4; }}
	private final class OpCB0x68 extends Opcode { // BIT  5,B 
		public final int exec() {
			bit(5, BC.getH());
			return 4; }}
	private final class OpCB0x69 extends Opcode { // BIT  5,C 
		public final int exec() {
			bit(5, BC.getL());
			return 4; }}
	private final class OpCB0x6A extends Opcode { // BIT  5,D 
		public final int exec() {
			bit(5, DE.getH());
			return 4; }}
	private final class OpCB0x6B extends Opcode { // BIT  5,E 
		public final int exec() {
			bit(5, DE.getL());
			return 4; }}
	private final class OpCB0x6C extends Opcode { // BIT  5,H 
		public final int exec() {
			bit(5, HL.getH());
			return 4; }}
	private final class OpCB0x6D extends Opcode { // BIT  5,L 
		public final int exec() {
			bit(5, HL.getL());
			return 4; }}
	private final class OpCB0x6E extends Opcode { // BIT  5,(HL) 
		public final int exec() {
			bit(5, HL.getI(mem));
			return 4; }}
	private final class OpCB0x6F extends Opcode { // BIT  5,A 
		public final int exec() {
			bit(5, AF.getH());
			return 4; }}
	private final class OpCB0x70 extends Opcode { // BIT  6,B 
		public final int exec() {
			bit(6, BC.getH());
			return 4; }}
	private final class OpCB0x71 extends Opcode { // BIT  6,C 
		public final int exec() {
			bit(6, BC.getL());
			return 4; }}
	private final class OpCB0x72 extends Opcode { // BIT  6,D 
		public final int exec() {
			bit(6, DE.getH());
			return 4; }}
	private final class OpCB0x73 extends Opcode { // BIT  6,E 
		public final int exec() {
			bit(6, DE.getL());
			return 4; }}
	private final class OpCB0x74 extends Opcode { // BIT  6,H 
		public final int exec() {
			bit(6, HL.getH());
			return 4; }}
	private final class OpCB0x75 extends Opcode { // BIT  6,L 
		public final int exec() {
			bit(6, HL.getL());
			return 4; }}
	private final class OpCB0x76 extends Opcode { // BIT  6,(HL) 
		public final int exec() {
			bit(6, HL.getI(mem));
			return 4; }}
	private final class OpCB0x77 extends Opcode { // BIT  6,A 
		public final int exec() {
			bit(6, AF.getH());
			return 4; }}
	private final class OpCB0x78 extends Opcode { // BIT  7,B 
		public final int exec() {
			bit(7, BC.getH());
			return 4; }}
	private final class OpCB0x79 extends Opcode { // BIT  7,C 
		public final int exec() {
			bit(7, BC.getL());
			return 4; }}
	private final class OpCB0x7A extends Opcode { // BIT  7,D 
		public final int exec() {
			bit(7, DE.getH());
			return 4; }}
	private final class OpCB0x7B extends Opcode { // BIT  7,E 
		public final int exec() {
			bit(7, DE.getL());
			return 4; }}
	private final class OpCB0x7C extends Opcode { // BIT  7,H 
		public final int exec() {
			bit(7, HL.getH());
			return 4; }}
	private final class OpCB0x7D extends Opcode { // BIT  7,L
		public final int exec() {
			bit(7, HL.getL());
			return 4; }}
	private final class OpCB0x7E extends Opcode { // BIT  7,(HL)
		public final int exec() {
			bit(7, HL.getI(mem));
			return 4; }}
	private final class OpCB0x7F extends Opcode { // BIT  7,A
		public final int exec() {
			bit(7, AF.getH());
			return 4; }}
	private final class OpCB0x80 extends Opcode { // RES 0, B
		public final int exec() {
			resH(0, BC);
			return 4; }}
	private final class OpCB0x81 extends Opcode { // RES 0, C
		public final int exec() {
			resL(0, BC);
			return 4; }}
	private final class OpCB0x82 extends Opcode { // RES 0, D
		public final int exec() {
			resH(0, DE);
			return 4; }}
	private final class OpCB0x83 extends Opcode { // RES 0, E
		public final int exec() {
			resL(0, DE);
			return 4; }}
	private final class OpCB0x84 extends Opcode { // RES 0, H
		public final int exec() {
			resH(0, HL);
			return 4; }}
	private final class OpCB0x85 extends Opcode { // RES 0, L
		public final int exec() {
			resL(0, HL);
			return 4; }}
	private final class OpCB0x86 extends Opcode { // RES 0, (HL)
		public final int exec() {
			resI(0, HL.getR());
			return 4; }}
	private final class OpCB0x87 extends Opcode { // RES 0, A
		public final int exec() {
			resH(0, AF);
			return 4; }}
	private final class OpCB0x88 extends Opcode { // RES 1, B
		public final int exec() {
			resH(1, BC);
			return 4; }}
	private final class OpCB0x89 extends Opcode { // RES 1, C
		public final int exec() {
			resL(1, BC);
			return 4; }}
	private final class OpCB0x8A extends Opcode { // RES 1, D
		public final int exec() {
			resH(1, DE);
			return 4; }}
	private final class OpCB0x8B extends Opcode { // RES 1, E
		public final int exec() {
			resL(1, DE);
			return 4; }}
	private final class OpCB0x8C extends Opcode { // RES 1, H
		public final int exec() {
			resH(1, HL);
			return 4; }}
	private final class OpCB0x8D extends Opcode { // RES 1, L
		public final int exec() {
			resL(1, HL);
			return 4; }}
	private final class OpCB0x8E extends Opcode { // RES 1, (HL)
		public final int exec() {
			resI(1, HL.getR());
			return 4; }}
	private final class OpCB0x8F extends Opcode { // RES 1, A
		public final int exec() {
			resH(1, AF);
			return 4; }}
	private final class OpCB0x90 extends Opcode { // RES 2, B
		public final int exec() {
			resH(2, BC);
			return 4; }}
	private final class OpCB0x91 extends Opcode { // RES 2, C
		public final int exec() {
			resL(2, BC);
			return 4; }}
	private final class OpCB0x92 extends Opcode { // RES 2, D
		public final int exec() {
			resH(2, DE);
			return 4; }}
	private final class OpCB0x93 extends Opcode { // RES 2, E
		public final int exec() {
			resL(2, DE);
			return 4; }}
	private final class OpCB0x94 extends Opcode { // RES 2, H
		public final int exec() {
			resH(2, HL);
			return 4; }}
	private final class OpCB0x95 extends Opcode { // RES 2, L
		public final int exec() {
			resL(2, HL);
			return 4; }}
	private final class OpCB0x96 extends Opcode { // RES 2, (HL)
		public final int exec() {
			resI(2, HL.getR());
			return 4; }}
	private final class OpCB0x97 extends Opcode { // RES 2, A
		public final int exec() {
			resH(2, AF);
			return 4; }}
	private final class OpCB0x98 extends Opcode { // RES 3, B
		public final int exec() {
			resH(3, BC);
			return 4; }}
	private final class OpCB0x99 extends Opcode { // RES 3, C
		public final int exec() {
			resL(3, BC);
			return 4; }}
	private final class OpCB0x9A extends Opcode { // RES 3, D
		public final int exec() {
			resH(3, DE);
			return 4; }}
	private final class OpCB0x9B extends Opcode { // RES 3, E
		public final int exec() {
			resL(3, DE);
			return 4; }}
	private final class OpCB0x9C extends Opcode { // RES 3, H
		public final int exec() {
			resH(3, HL);
			return 4; }}
	private final class OpCB0x9D extends Opcode { // RES 3, L
		public final int exec() {
			resL(3, HL);
			return 4; }}
	private final class OpCB0x9E extends Opcode { // RES 3, (HL)
		public final int exec() {
			resI(3, HL.getR());
			return 4; }}
	private final class OpCB0x9F extends Opcode { // RES 3, A
		public final int exec() {
			resH(3, AF);
			return 4; }}
	private final class OpCB0xA0 extends Opcode { // RES 4, B
		public final int exec() {
			resH(4, BC);
			return 4; }}
	private final class OpCB0xA1 extends Opcode { // RES 4, C
		public final int exec() {
			resL(4, BC);
			return 4; }}
	private final class OpCB0xA2 extends Opcode { // RES 4, D
		public final int exec() {
			resH(4, DE);
			return 4; }}
	private final class OpCB0xA3 extends Opcode { // RES 4, E
		public final int exec() {
			resL(4, DE);
			return 4; }}
	private final class OpCB0xA4 extends Opcode { // RES 4, H
		public final int exec() {
			resH(4, HL);
			return 4; }}
	private final class OpCB0xA5 extends Opcode { // RES 4, L
		public final int exec() {
			resL(4, HL);
			return 4; }}
	private final class OpCB0xA6 extends Opcode { // RES 4, (HL)
		public final int exec() {
			resI(4, HL.getR());
			return 4; }}
	private final class OpCB0xA7 extends Opcode { // RES 4, A
		public final int exec() {
			resH(4, AF);
			return 4; }}
	private final class OpCB0xA8 extends Opcode { // RES 5, B
		public final int exec() {
			resH(5, BC);
			return 4; }}
	private final class OpCB0xA9 extends Opcode { // RES 5, C
		public final int exec() {
			resL(5, BC);
			return 4; }}
	private final class OpCB0xAA extends Opcode { // RES 5, D
		public final int exec() {
			resH(5, DE);
			return 4; }}
	private final class OpCB0xAB extends Opcode { // RES 5, E
		public final int exec() {
			resL(5, DE);
			return 4; }}
	private final class OpCB0xAC extends Opcode { // RES 5, H
		public final int exec() {
			resH(5, HL);
			return 4; }}
	private final class OpCB0xAD extends Opcode { // RES 5, L
		public final int exec() {
			resL(5, HL);
			return 4; }}
	private final class OpCB0xAE extends Opcode { // RES 5, (HL)
		public final int exec() {
			resI(5, HL.getR());
			return 4; }}
	private final class OpCB0xAF extends Opcode { // RES 5, A
		public final int exec() {
			resH(5, AF);
			return 4; }}
	private final class OpCB0xB0 extends Opcode { // RES 6, B
		public final int exec() {
			resH(6, BC);
			return 4; }}
	private final class OpCB0xB1 extends Opcode { // RES 6, C
		public final int exec() {
			resL(6, BC);
			return 4; }}
	private final class OpCB0xB2 extends Opcode { // RES 6, D
		public final int exec() {
			resH(6, DE);
			return 4; }}
	private final class OpCB0xB3 extends Opcode { // RES 6, E
		public final int exec() {
			resL(6, DE);
			return 4; }}
	private final class OpCB0xB4 extends Opcode { // RES 6, H
		public final int exec() {
			resH(6, HL);
			return 4; }}
	private final class OpCB0xB5 extends Opcode { // RES 6, L
		public final int exec() {
			resL(6, HL);
			return 4; }}
	private final class OpCB0xB6 extends Opcode { // RES 6, (HL)
		public final int exec() {
			resI(6, HL.getR());
			return 4; }}
	private final class OpCB0xB7 extends Opcode { // RES 6, A
		public final int exec() {
			resH(6, AF);
			return 4; }}
	private final class OpCB0xB8 extends Opcode { // RES 7, B
		public final int exec() {
			resH(7, BC);
			return 4; }}
	private final class OpCB0xB9 extends Opcode { // RES 7, C
		public final int exec() {
			resL(7, BC);
			return 4; }}
	private final class OpCB0xBA extends Opcode { // RES 7, D
		public final int exec() {
			resH(7, DE);
			return 4; }}
	private final class OpCB0xBB extends Opcode { // RES 7, E
		public final int exec() {
			resL(7, DE);
			return 4; }}
	private final class OpCB0xBC extends Opcode { // RES 7, H
		public final int exec() {
			resH(7, HL);
			return 4; }}
	private final class OpCB0xBD extends Opcode { // RES 7, L
		public final int exec() {
			resL(7, HL);
			return 4; }}
	private final class OpCB0xBE extends Opcode { // RES 7, (HL)
		public final int exec() {
			resI(7, HL.getR());
			return 4; }}
	private final class OpCB0xBF extends Opcode { // RES 7, A
		public final int exec() {
			resH(7, AF);
			return 4; }}
	private final class OpCB0xC0 extends Opcode { // SET 0, B
		public final int exec() {
			setH(0, BC);
			return 4; }}
	private final class OpCB0xC1 extends Opcode { // SET 0, C
		public final int exec() {
			setL(0, BC);
			return 4; }}
	private final class OpCB0xC2 extends Opcode { // SET 0, D
		public final int exec() {
			setH(0, DE);
			return 4; }}
	private final class OpCB0xC3 extends Opcode { // SET 0, E
		public final int exec() {
			setL(0, DE);
			return 4; }}
	private final class OpCB0xC4 extends Opcode { // SET 0, H
		public final int exec() {
			setH(0, HL);
			return 4; }}
	private final class OpCB0xC5 extends Opcode { // SET 0, L
		public final int exec() {
			setL(0, HL);
			return 4; }}
	private final class OpCB0xC6 extends Opcode { // SET 0, (HL)
		public final int exec() {
			setI(0, HL.getR());
			return 4; }}
	private final class OpCB0xC7 extends Opcode { // SET 0, A
		public final int exec() {
			setH(0, AF);
			return 4; }}
	private final class OpCB0xC8 extends Opcode { // SET 1, B
		public final int exec() {
			setH(1, BC);
			return 4; }}
	private final class OpCB0xC9 extends Opcode { // SET 1, C
		public final int exec() {
			setL(1, BC);
			return 4; }}
	private final class OpCB0xCA extends Opcode { // SET 1, D
		public final int exec() {
			setH(1, DE);
			return 4; }}
	private final class OpCB0xCB extends Opcode { // SET 1, E
		public final int exec() {
			setL(1, DE);
			return 4; }}
	private final class OpCB0xCC extends Opcode { // SET 1, H
		public final int exec() {
			setH(1, HL);
			return 4; }}
	private final class OpCB0xCD extends Opcode { // SET 1, L
		public final int exec() {
			setL(1, HL);
			return 4; }}
	private final class OpCB0xCE extends Opcode { // SET 1, (HL)
		public final int exec() {
			setI(1, HL.getR());
			return 4; }}
	private final class OpCB0xCF extends Opcode { // SET 1, A
		public final int exec() {
			setH(1, AF);
			return 4; }}
	private final class OpCB0xD0 extends Opcode { // SET 2, B
		public final int exec() {
			setH(2, BC);
			return 4; }}
	private final class OpCB0xD1 extends Opcode { // SET 2, C
		public final int exec() {
			setL(2, BC);
			return 4; }}
	private final class OpCB0xD2 extends Opcode { // SET 2, D
		public final int exec() {
			setH(2, DE);
			return 4; }}
	private final class OpCB0xD3 extends Opcode { // SET 2, E
		public final int exec() {
			setL(2, DE);
			return 4; }}
	private final class OpCB0xD4 extends Opcode { // SET 2, H
		public final int exec() {
			setH(2, HL);
			return 4; }}
	private final class OpCB0xD5 extends Opcode { // SET 2, L
		public final int exec() {
			setL(2, HL);
			return 4; }}
	private final class OpCB0xD6 extends Opcode { // SET 2, (HL)
		public final int exec() {
			setI(2, HL.getR());
			return 4; }}
	private final class OpCB0xD7 extends Opcode { // SET 2, A
		public final int exec() {
			setH(2, AF);
			return 4; }}
	private final class OpCB0xD8 extends Opcode { // SET 3, B
		public final int exec() {
			setH(3, BC);
			return 4; }}
	private final class OpCB0xD9 extends Opcode { // SET 3, C
		public final int exec() {
			setL(3, BC);
			return 4; }}
	private final class OpCB0xDA extends Opcode { // SET 3, D
		public final int exec() {
			setH(3, DE);
			return 4; }}
	private final class OpCB0xDB extends Opcode { // SET 3, E
		public final int exec() {
			setL(3, DE);
			return 4; }}
	private final class OpCB0xDC extends Opcode { // SET 3, H
		public final int exec() {
			setH(3, HL);
			return 4; }}
	private final class OpCB0xDD extends Opcode { // SET 3, L
		public final int exec() {
			setL(3, HL);
			return 4; }}
	private final class OpCB0xDE extends Opcode { // SET 3, (HL)
		public final int exec() {
			setI(3, HL.getR());
			return 4; }}
	private final class OpCB0xDF extends Opcode { // SET 3, A
		public final int exec() {
			setH(3, AF);
			return 4; }}
	private final class OpCB0xE0 extends Opcode { // SET 4, B
		public final int exec() {
			setH(4, BC);
			return 4; }}
	private final class OpCB0xE1 extends Opcode { // SET 4, C
		public final int exec() {
			setL(4, BC);
			return 4; }}
	private final class OpCB0xE2 extends Opcode { // SET 4, D
		public final int exec() {
			setH(4, DE);
			return 4; }}
	private final class OpCB0xE3 extends Opcode { // SET 4, E
		public final int exec() {
			setL(4, DE);
			return 4; }}
	private final class OpCB0xE4 extends Opcode { // SET 4, H
		public final int exec() {
			setH(4, HL);
			return 4; }}
	private final class OpCB0xE5 extends Opcode { // SET 4, L
		public final int exec() {
			setL(4, HL);
			return 4; }}
	private final class OpCB0xE6 extends Opcode { // SET 4, (HL)
		public final int exec() {
			setI(4, HL.getR());
			return 4; }}
	private final class OpCB0xE7 extends Opcode { // SET 4, A
		public final int exec() {
			setH(4, AF);
			return 4; }}
	private final class OpCB0xE8 extends Opcode { // SET 5, B
		public final int exec() {
			setH(5, BC);
			return 4; }}
	private final class OpCB0xE9 extends Opcode { // SET 5, C
		public final int exec() {
			setL(5, BC);
			return 4; }}
	private final class OpCB0xEA extends Opcode { // SET 5, D
		public final int exec() {
			setH(5, DE);
			return 4; }}
	private final class OpCB0xEB extends Opcode { // SET 5, E
		public final int exec() {
			setL(5, DE);
			return 4; }}
	private final class OpCB0xEC extends Opcode { // SET 5, H
		public final int exec() {
			setH(5, HL);
			return 4; }}
	private final class OpCB0xED extends Opcode { // SET 5, L
		public final int exec() {
			setL(5, HL);
			return 4; }}
	private final class OpCB0xEE extends Opcode { // SET 5, (HL)
		public final int exec() {
			setI(5, HL.getR());
			return 4; }}
	private final class OpCB0xEF extends Opcode { // SET 5, A
		public final int exec() {
			setH(5, AF);
			return 4; }}
	private final class OpCB0xF0 extends Opcode { // SET 6, B
		public final int exec() {
			setH(6, BC);
			return 4; }}
	private final class OpCB0xF1 extends Opcode { // SET 6, C
		public final int exec() {
			setL(6, BC);
			return 4; }}
	private final class OpCB0xF2 extends Opcode { // SET 6, D
		public final int exec() {
			setH(6, DE);
			return 4; }}
	private final class OpCB0xF3 extends Opcode { // SET 6, E
		public final int exec() {
			setL(6, DE);
			return 4; }}
	private final class OpCB0xF4 extends Opcode { // SET 6, H
		public final int exec() {
			setH(6, HL);
			return 4; }}
	private final class OpCB0xF5 extends Opcode { // SET 6, L
		public final int exec() {
			setL(6, HL);
			return 4; }}
	private final class OpCB0xF6 extends Opcode { // SET 6, (HL)
		public final int exec() {
			setI(6, HL.getR());
			return 4; }}
	private final class OpCB0xF7 extends Opcode { // SET 6, A
		public final int exec() {
			setH(6, AF);
			return 4; }}
	private final class OpCB0xF8 extends Opcode { // SET 7, B
		public final int exec() {
			setH(7, BC);
			return 4; }}
	private final class OpCB0xF9 extends Opcode { // SET 7, C
		public final int exec() {
			setL(7, BC);
			return 4; }}
	private final class OpCB0xFA extends Opcode { // SET 7, D
		public final int exec() {
			setH(7, DE);
			return 4; }}
	private final class OpCB0xFB extends Opcode { // SET 7, E
		public final int exec() {
			setL(7, DE);
			return 4; }}
	private final class OpCB0xFC extends Opcode { // SET 7, H
		public final int exec() {
			setH(7, HL);
			return 4; }}
	private final class OpCB0xFD extends Opcode { // SET 7, L
		public final int exec() {
			setL(7, HL);
			return 4; }}
	private final class OpCB0xFE extends Opcode { // SET 7, (HL)
		public final int exec() {
			setI(7, HL.getR());
			return 4; }}
	private final class OpCB0xFF extends Opcode { // SET 7, A
		public final int exec() {
			setH(7, AF);
			return 4; }}
	
	/**
	 * init opcodes
	 */
	private void initOpcodes() {
		opcodes = new Opcode[256];
		
		opcodes[0x00] = new Op0x00();
		opcodes[0x01] = new Op0x01();
		opcodes[0x02] = new Op0x02();
		opcodes[0x03] = new Op0x03();
		opcodes[0x04] = new Op0x04();
		opcodes[0x05] = new Op0x05();
		opcodes[0x06] = new Op0x06();
		opcodes[0x07] = new Op0x07();
		opcodes[0x08] = new Op0x08();
		opcodes[0x09] = new Op0x09();
		opcodes[0x0A] = new Op0x0A();
		opcodes[0x0B] = new Op0x0B();
		opcodes[0x0C] = new Op0x0C();
		opcodes[0x0D] = new Op0x0D();
		opcodes[0x0E] = new Op0x0E();
		opcodes[0x0F] = new Op0x0F();
		opcodes[0x10] = new Op0x10();
		opcodes[0x11] = new Op0x11();
		opcodes[0x12] = new Op0x12();
		opcodes[0x13] = new Op0x13();
		opcodes[0x14] = new Op0x14();
		opcodes[0x15] = new Op0x15();
		opcodes[0x16] = new Op0x16();
		opcodes[0x17] = new Op0x17();
		opcodes[0x18] = new Op0x18();
		opcodes[0x19] = new Op0x19();
		opcodes[0x1A] = new Op0x1A();
		opcodes[0x1B] = new Op0x1B();
		opcodes[0x1C] = new Op0x1C();
		opcodes[0x1D] = new Op0x1D();
		opcodes[0x1E] = new Op0x1E();
		opcodes[0x1F] = new Op0x1F();
		opcodes[0x20] = new Op0x20();
		opcodes[0x21] = new Op0x21();
		opcodes[0x22] = new Op0x22();
		opcodes[0x23] = new Op0x23();
		opcodes[0x24] = new Op0x24();
		opcodes[0x25] = new Op0x25();
		opcodes[0x26] = new Op0x26();
		opcodes[0x27] = new Op0x27();
		opcodes[0x28] = new Op0x28();
		opcodes[0x29] = new Op0x29();
		opcodes[0x2A] = new Op0x2A();
		opcodes[0x2B] = new Op0x2B();
		opcodes[0x2C] = new Op0x2C();
		opcodes[0x2D] = new Op0x2D();
		opcodes[0x2E] = new Op0x2E();
		opcodes[0x2F] = new Op0x2F();
		opcodes[0x30] = new Op0x30();
		opcodes[0x31] = new Op0x31();
		opcodes[0x32] = new Op0x32();
		opcodes[0x33] = new Op0x33();
		opcodes[0x34] = new Op0x34();
		opcodes[0x35] = new Op0x35();
		opcodes[0x36] = new Op0x36();
		opcodes[0x37] = new Op0x37();
		opcodes[0x38] = new Op0x38();
		opcodes[0x39] = new Op0x39();
		opcodes[0x3A] = new Op0x3A();
		opcodes[0x3B] = new Op0x3B();
		opcodes[0x3C] = new Op0x3C();
		opcodes[0x3D] = new Op0x3D();
		opcodes[0x3E] = new Op0x3E();
		opcodes[0x3F] = new Op0x3F();
		opcodes[0x40] = new Op0x40();
		opcodes[0x41] = new Op0x41();
		opcodes[0x42] = new Op0x42();
		opcodes[0x43] = new Op0x43();
		opcodes[0x44] = new Op0x44();
		opcodes[0x45] = new Op0x45();
		opcodes[0x46] = new Op0x46();
		opcodes[0x47] = new Op0x47();
		opcodes[0x48] = new Op0x48();
		opcodes[0x49] = new Op0x49();
		opcodes[0x4A] = new Op0x4A();
		opcodes[0x4B] = new Op0x4B();
		opcodes[0x4C] = new Op0x4C();
		opcodes[0x4D] = new Op0x4D();
		opcodes[0x4E] = new Op0x4E();
		opcodes[0x4F] = new Op0x4F();
		opcodes[0x50] = new Op0x50();
		opcodes[0x51] = new Op0x51();
		opcodes[0x52] = new Op0x52();
		opcodes[0x53] = new Op0x53();
		opcodes[0x54] = new Op0x54();
		opcodes[0x55] = new Op0x55();
		opcodes[0x56] = new Op0x56();
		opcodes[0x57] = new Op0x57();
		opcodes[0x58] = new Op0x58();
		opcodes[0x59] = new Op0x59();
		opcodes[0x5A] = new Op0x5A();
		opcodes[0x5B] = new Op0x5B();
		opcodes[0x5C] = new Op0x5C();
		opcodes[0x5D] = new Op0x5D();
		opcodes[0x5E] = new Op0x5E();
		opcodes[0x5F] = new Op0x5F();
		opcodes[0x60] = new Op0x60();
		opcodes[0x61] = new Op0x61();
		opcodes[0x62] = new Op0x62();
		opcodes[0x63] = new Op0x63();
		opcodes[0x64] = new Op0x64();
		opcodes[0x65] = new Op0x65();
		opcodes[0x66] = new Op0x66();
		opcodes[0x67] = new Op0x67();
		opcodes[0x68] = new Op0x68();
		opcodes[0x69] = new Op0x69();
		opcodes[0x6A] = new Op0x6A();
		opcodes[0x6B] = new Op0x6B();
		opcodes[0x6C] = new Op0x6C();
		opcodes[0x6D] = new Op0x6D();
		opcodes[0x6E] = new Op0x6E();
		opcodes[0x6F] = new Op0x6F();
		opcodes[0x70] = new Op0x70();
		opcodes[0x71] = new Op0x71();
		opcodes[0x72] = new Op0x72();
		opcodes[0x73] = new Op0x73();
		opcodes[0x74] = new Op0x74();
		opcodes[0x75] = new Op0x75();
		opcodes[0x76] = new Op0x76();
		opcodes[0x77] = new Op0x77();
		opcodes[0x78] = new Op0x78();
		opcodes[0x79] = new Op0x79();
		opcodes[0x7A] = new Op0x7A();
		opcodes[0x7B] = new Op0x7B();
		opcodes[0x7C] = new Op0x7C();
		opcodes[0x7D] = new Op0x7D();
		opcodes[0x7E] = new Op0x7E();
		opcodes[0x7F] = new Op0x7F();
		opcodes[0x80] = new Op0x80();
		opcodes[0x81] = new Op0x81();
		opcodes[0x82] = new Op0x82();
		opcodes[0x83] = new Op0x83();
		opcodes[0x84] = new Op0x84();
		opcodes[0x85] = new Op0x85();
		opcodes[0x86] = new Op0x86();
		opcodes[0x87] = new Op0x87();
		opcodes[0x88] = new Op0x88();
		opcodes[0x89] = new Op0x89();
		opcodes[0x8A] = new Op0x8A();
		opcodes[0x8B] = new Op0x8B();
		opcodes[0x8C] = new Op0x8C();
		opcodes[0x8D] = new Op0x8D();
		opcodes[0x8E] = new Op0x8E();
		opcodes[0x8F] = new Op0x8F();
		opcodes[0x90] = new Op0x90();
		opcodes[0x91] = new Op0x91();
		opcodes[0x92] = new Op0x92();
		opcodes[0x93] = new Op0x93();
		opcodes[0x94] = new Op0x94();
		opcodes[0x95] = new Op0x95();
		opcodes[0x96] = new Op0x96();
		opcodes[0x97] = new Op0x97();
		opcodes[0x98] = new Op0x98();
		opcodes[0x99] = new Op0x99();
		opcodes[0x9A] = new Op0x9A();
		opcodes[0x9B] = new Op0x9B();
		opcodes[0x9C] = new Op0x9C();
		opcodes[0x9D] = new Op0x9D();
		opcodes[0x9E] = new Op0x9E();
		opcodes[0x9F] = new Op0x9F();
		opcodes[0xA0] = new Op0xA0();
		opcodes[0xA1] = new Op0xA1();
		opcodes[0xA2] = new Op0xA2();
		opcodes[0xA3] = new Op0xA3();
		opcodes[0xA4] = new Op0xA4();
		opcodes[0xA5] = new Op0xA5();
		opcodes[0xA6] = new Op0xA6();
		opcodes[0xA7] = new Op0xA7();
		opcodes[0xA8] = new Op0xA8();
		opcodes[0xA9] = new Op0xA9();
		opcodes[0xAA] = new Op0xAA();
		opcodes[0xAB] = new Op0xAB();
		opcodes[0xAC] = new Op0xAC();
		opcodes[0xAD] = new Op0xAD();
		opcodes[0xAE] = new Op0xAE();
		opcodes[0xAF] = new Op0xAF();
		opcodes[0xB0] = new Op0xB0();
		opcodes[0xB1] = new Op0xB1();
		opcodes[0xB2] = new Op0xB2();
		opcodes[0xB3] = new Op0xB3();
		opcodes[0xB4] = new Op0xB4();
		opcodes[0xB5] = new Op0xB5();
		opcodes[0xB6] = new Op0xB6();
		opcodes[0xB7] = new Op0xB7();
		opcodes[0xB8] = new Op0xB8();
		opcodes[0xB9] = new Op0xB9();
		opcodes[0xBA] = new Op0xBA();
		opcodes[0xBB] = new Op0xBB();
		opcodes[0xBC] = new Op0xBC();
		opcodes[0xBD] = new Op0xBD();
		opcodes[0xBE] = new Op0xBE();
		opcodes[0xBF] = new Op0xBF();
		opcodes[0xC0] = new Op0xC0();
		opcodes[0xC1] = new Op0xC1();
		opcodes[0xC2] = new Op0xC2();
		opcodes[0xC3] = new Op0xC3();
		opcodes[0xC4] = new Op0xC4();
		opcodes[0xC5] = new Op0xC5();
		opcodes[0xC6] = new Op0xC6();
		opcodes[0xC7] = new Op0xC7();
		opcodes[0xC8] = new Op0xC8();
		opcodes[0xC9] = new Op0xC9();
		opcodes[0xCA] = new Op0xCA();
		opcodes[0xCB] = new Op0xCB();
		opcodes[0xCC] = new Op0xCC();
		opcodes[0xCD] = new Op0xCD();
		opcodes[0xCE] = new Op0xCE();
		opcodes[0xCF] = new Op0xCF();
		opcodes[0xD0] = new Op0xD0();
		opcodes[0xD1] = new Op0xD1();
		opcodes[0xD2] = new Op0xD2();
		opcodes[0xD3] = new Op0xD3();
		opcodes[0xD4] = new Op0xD4();
		opcodes[0xD5] = new Op0xD5();
		opcodes[0xD6] = new Op0xD6();
		opcodes[0xD7] = new Op0xD7();
		opcodes[0xD8] = new Op0xD8();
		opcodes[0xD9] = new Op0xD9();
		opcodes[0xDA] = new Op0xDA();
		opcodes[0xDB] = new Op0xDB();
		opcodes[0xDC] = new Op0xDC();
		opcodes[0xDD] = new Op0xDD();
		opcodes[0xDE] = new Op0xDE();
		opcodes[0xDF] = new Op0xDF();
		opcodes[0xE0] = new Op0xE0();
		opcodes[0xE1] = new Op0xE1();
		opcodes[0xE2] = new Op0xE2();
		opcodes[0xE3] = new Op0xE3();
		opcodes[0xE4] = new Op0xE4();
		opcodes[0xE5] = new Op0xE5();
		opcodes[0xE6] = new Op0xE6();
		opcodes[0xE7] = new Op0xE7();
		opcodes[0xE8] = new Op0xE8();
		opcodes[0xE9] = new Op0xE9();
		opcodes[0xEA] = new Op0xEA();
		opcodes[0xEB] = new Op0xEB();
		opcodes[0xEC] = new Op0xEC();
		opcodes[0xED] = new Op0xED();
		opcodes[0xEE] = new Op0xEE();
		opcodes[0xEF] = new Op0xEF();
		opcodes[0xF0] = new Op0xF0();
		opcodes[0xF1] = new Op0xF1();
		opcodes[0xF2] = new Op0xF2();
		opcodes[0xF3] = new Op0xF3();
		opcodes[0xF4] = new Op0xF4();
		opcodes[0xF5] = new Op0xF5();
		opcodes[0xF6] = new Op0xF6();
		opcodes[0xF7] = new Op0xF7();
		opcodes[0xF8] = new Op0xF8();
		opcodes[0xF9] = new Op0xF9();
		opcodes[0xFA] = new Op0xFA();
		opcodes[0xFB] = new Op0xFB();
		opcodes[0xFC] = new Op0xFC();
		opcodes[0xFD] = new Op0xFD();
		opcodes[0xFE] = new Op0xFE();
		opcodes[0xFF] = new Op0xFF();
		
		cbcodes = new Opcode[256];
		
		cbcodes[0x00] = new OpCB0x00();
		cbcodes[0x01] = new OpCB0x01();
		cbcodes[0x02] = new OpCB0x02();
		cbcodes[0x03] = new OpCB0x03();
		cbcodes[0x04] = new OpCB0x04();
		cbcodes[0x05] = new OpCB0x05();
		cbcodes[0x06] = new OpCB0x06();
		cbcodes[0x07] = new OpCB0x07();
		cbcodes[0x08] = new OpCB0x08();
		cbcodes[0x09] = new OpCB0x09();
		cbcodes[0x0A] = new OpCB0x0A();
		cbcodes[0x0B] = new OpCB0x0B();
		cbcodes[0x0C] = new OpCB0x0C();
		cbcodes[0x0D] = new OpCB0x0D();
		cbcodes[0x0E] = new OpCB0x0E();
		cbcodes[0x0F] = new OpCB0x0F();
		cbcodes[0x10] = new OpCB0x10();
		cbcodes[0x11] = new OpCB0x11();
		cbcodes[0x12] = new OpCB0x12();
		cbcodes[0x13] = new OpCB0x13();
		cbcodes[0x14] = new OpCB0x14();
		cbcodes[0x15] = new OpCB0x15();
		cbcodes[0x16] = new OpCB0x16();
		cbcodes[0x17] = new OpCB0x17();
		cbcodes[0x18] = new OpCB0x18();
		cbcodes[0x19] = new OpCB0x19();
		cbcodes[0x1A] = new OpCB0x1A();
		cbcodes[0x1B] = new OpCB0x1B();
		cbcodes[0x1C] = new OpCB0x1C();
		cbcodes[0x1D] = new OpCB0x1D();
		cbcodes[0x1E] = new OpCB0x1E();
		cbcodes[0x1F] = new OpCB0x1F();
		cbcodes[0x20] = new OpCB0x20();
		cbcodes[0x21] = new OpCB0x21();
		cbcodes[0x22] = new OpCB0x22();
		cbcodes[0x23] = new OpCB0x23();
		cbcodes[0x24] = new OpCB0x24();
		cbcodes[0x25] = new OpCB0x25();
		cbcodes[0x26] = new OpCB0x26();
		cbcodes[0x27] = new OpCB0x27();
		cbcodes[0x28] = new OpCB0x28();
		cbcodes[0x29] = new OpCB0x29();
		cbcodes[0x2A] = new OpCB0x2A();
		cbcodes[0x2B] = new OpCB0x2B();
		cbcodes[0x2C] = new OpCB0x2C();
		cbcodes[0x2D] = new OpCB0x2D();
		cbcodes[0x2E] = new OpCB0x2E();
		cbcodes[0x2F] = new OpCB0x2F();
		cbcodes[0x30] = new OpCB0x30();
		cbcodes[0x31] = new OpCB0x31();
		cbcodes[0x32] = new OpCB0x32();
		cbcodes[0x33] = new OpCB0x33();
		cbcodes[0x34] = new OpCB0x34();
		cbcodes[0x35] = new OpCB0x35();
		cbcodes[0x36] = new OpCB0x36();
		cbcodes[0x37] = new OpCB0x37();
		cbcodes[0x38] = new OpCB0x38();
		cbcodes[0x39] = new OpCB0x39();
		cbcodes[0x3A] = new OpCB0x3A();
		cbcodes[0x3B] = new OpCB0x3B();
		cbcodes[0x3C] = new OpCB0x3C();
		cbcodes[0x3D] = new OpCB0x3D();
		cbcodes[0x3E] = new OpCB0x3E();
		cbcodes[0x3F] = new OpCB0x3F();
		cbcodes[0x40] = new OpCB0x40();
		cbcodes[0x41] = new OpCB0x41();
		cbcodes[0x42] = new OpCB0x42();
		cbcodes[0x43] = new OpCB0x43();
		cbcodes[0x44] = new OpCB0x44();
		cbcodes[0x45] = new OpCB0x45();
		cbcodes[0x46] = new OpCB0x46();
		cbcodes[0x47] = new OpCB0x47();
		cbcodes[0x48] = new OpCB0x48();
		cbcodes[0x49] = new OpCB0x49();
		cbcodes[0x4A] = new OpCB0x4A();
		cbcodes[0x4B] = new OpCB0x4B();
		cbcodes[0x4C] = new OpCB0x4C();
		cbcodes[0x4D] = new OpCB0x4D();
		cbcodes[0x4E] = new OpCB0x4E();
		cbcodes[0x4F] = new OpCB0x4F();
		cbcodes[0x50] = new OpCB0x50();
		cbcodes[0x51] = new OpCB0x51();
		cbcodes[0x52] = new OpCB0x52();
		cbcodes[0x53] = new OpCB0x53();
		cbcodes[0x54] = new OpCB0x54();
		cbcodes[0x55] = new OpCB0x55();
		cbcodes[0x56] = new OpCB0x56();
		cbcodes[0x57] = new OpCB0x57();
		cbcodes[0x58] = new OpCB0x58();
		cbcodes[0x59] = new OpCB0x59();
		cbcodes[0x5A] = new OpCB0x5A();
		cbcodes[0x5B] = new OpCB0x5B();
		cbcodes[0x5C] = new OpCB0x5C();
		cbcodes[0x5D] = new OpCB0x5D();
		cbcodes[0x5E] = new OpCB0x5E();
		cbcodes[0x5F] = new OpCB0x5F();
		cbcodes[0x60] = new OpCB0x60();
		cbcodes[0x61] = new OpCB0x61();
		cbcodes[0x62] = new OpCB0x62();
		cbcodes[0x63] = new OpCB0x63();
		cbcodes[0x64] = new OpCB0x64();
		cbcodes[0x65] = new OpCB0x65();
		cbcodes[0x66] = new OpCB0x66();
		cbcodes[0x67] = new OpCB0x67();
		cbcodes[0x68] = new OpCB0x68();
		cbcodes[0x69] = new OpCB0x69();
		cbcodes[0x6A] = new OpCB0x6A();
		cbcodes[0x6B] = new OpCB0x6B();
		cbcodes[0x6C] = new OpCB0x6C();
		cbcodes[0x6D] = new OpCB0x6D();
		cbcodes[0x6E] = new OpCB0x6E();
		cbcodes[0x6F] = new OpCB0x6F();
		cbcodes[0x70] = new OpCB0x70();
		cbcodes[0x71] = new OpCB0x71();
		cbcodes[0x72] = new OpCB0x72();
		cbcodes[0x73] = new OpCB0x73();
		cbcodes[0x74] = new OpCB0x74();
		cbcodes[0x75] = new OpCB0x75();
		cbcodes[0x76] = new OpCB0x76();
		cbcodes[0x77] = new OpCB0x77();
		cbcodes[0x78] = new OpCB0x78();
		cbcodes[0x79] = new OpCB0x79();
		cbcodes[0x7A] = new OpCB0x7A();
		cbcodes[0x7B] = new OpCB0x7B();
		cbcodes[0x7C] = new OpCB0x7C();
		cbcodes[0x7D] = new OpCB0x7D();
		cbcodes[0x7E] = new OpCB0x7E();
		cbcodes[0x7F] = new OpCB0x7F();
		cbcodes[0x80] = new OpCB0x80();
		cbcodes[0x81] = new OpCB0x81();
		cbcodes[0x82] = new OpCB0x82();
		cbcodes[0x83] = new OpCB0x83();
		cbcodes[0x84] = new OpCB0x84();
		cbcodes[0x85] = new OpCB0x85();
		cbcodes[0x86] = new OpCB0x86();
		cbcodes[0x87] = new OpCB0x87();
		cbcodes[0x88] = new OpCB0x88();
		cbcodes[0x89] = new OpCB0x89();
		cbcodes[0x8A] = new OpCB0x8A();
		cbcodes[0x8B] = new OpCB0x8B();
		cbcodes[0x8C] = new OpCB0x8C();
		cbcodes[0x8D] = new OpCB0x8D();
		cbcodes[0x8E] = new OpCB0x8E();
		cbcodes[0x8F] = new OpCB0x8F();
		cbcodes[0x90] = new OpCB0x90();
		cbcodes[0x91] = new OpCB0x91();
		cbcodes[0x92] = new OpCB0x92();
		cbcodes[0x93] = new OpCB0x93();
		cbcodes[0x94] = new OpCB0x94();
		cbcodes[0x95] = new OpCB0x95();
		cbcodes[0x96] = new OpCB0x96();
		cbcodes[0x97] = new OpCB0x97();
		cbcodes[0x98] = new OpCB0x98();
		cbcodes[0x99] = new OpCB0x99();
		cbcodes[0x9A] = new OpCB0x9A();
		cbcodes[0x9B] = new OpCB0x9B();
		cbcodes[0x9C] = new OpCB0x9C();
		cbcodes[0x9D] = new OpCB0x9D();
		cbcodes[0x9E] = new OpCB0x9E();
		cbcodes[0x9F] = new OpCB0x9F();
		cbcodes[0xA0] = new OpCB0xA0();
		cbcodes[0xA1] = new OpCB0xA1();
		cbcodes[0xA2] = new OpCB0xA2();
		cbcodes[0xA3] = new OpCB0xA3();
		cbcodes[0xA4] = new OpCB0xA4();
		cbcodes[0xA5] = new OpCB0xA5();
		cbcodes[0xA6] = new OpCB0xA6();
		cbcodes[0xA7] = new OpCB0xA7();
		cbcodes[0xA8] = new OpCB0xA8();
		cbcodes[0xA9] = new OpCB0xA9();
		cbcodes[0xAA] = new OpCB0xAA();
		cbcodes[0xAB] = new OpCB0xAB();
		cbcodes[0xAC] = new OpCB0xAC();
		cbcodes[0xAD] = new OpCB0xAD();
		cbcodes[0xAE] = new OpCB0xAE();
		cbcodes[0xAF] = new OpCB0xAF();
		cbcodes[0xB0] = new OpCB0xB0();
		cbcodes[0xB1] = new OpCB0xB1();
		cbcodes[0xB2] = new OpCB0xB2();
		cbcodes[0xB3] = new OpCB0xB3();
		cbcodes[0xB4] = new OpCB0xB4();
		cbcodes[0xB5] = new OpCB0xB5();
		cbcodes[0xB6] = new OpCB0xB6();
		cbcodes[0xB7] = new OpCB0xB7();
		cbcodes[0xB8] = new OpCB0xB8();
		cbcodes[0xB9] = new OpCB0xB9();
		cbcodes[0xBA] = new OpCB0xBA();
		cbcodes[0xBB] = new OpCB0xBB();
		cbcodes[0xBC] = new OpCB0xBC();
		cbcodes[0xBD] = new OpCB0xBD();
		cbcodes[0xBE] = new OpCB0xBE();
		cbcodes[0xBF] = new OpCB0xBF();
		cbcodes[0xC0] = new OpCB0xC0();
		cbcodes[0xC1] = new OpCB0xC1();
		cbcodes[0xC2] = new OpCB0xC2();
		cbcodes[0xC3] = new OpCB0xC3();
		cbcodes[0xC4] = new OpCB0xC4();
		cbcodes[0xC5] = new OpCB0xC5();
		cbcodes[0xC6] = new OpCB0xC6();
		cbcodes[0xC7] = new OpCB0xC7();
		cbcodes[0xC8] = new OpCB0xC8();
		cbcodes[0xC9] = new OpCB0xC9();
		cbcodes[0xCA] = new OpCB0xCA();
		cbcodes[0xCB] = new OpCB0xCB();
		cbcodes[0xCC] = new OpCB0xCC();
		cbcodes[0xCD] = new OpCB0xCD();
		cbcodes[0xCE] = new OpCB0xCE();
		cbcodes[0xCF] = new OpCB0xCF();
		cbcodes[0xD0] = new OpCB0xD0();
		cbcodes[0xD1] = new OpCB0xD1();
		cbcodes[0xD2] = new OpCB0xD2();
		cbcodes[0xD3] = new OpCB0xD3();
		cbcodes[0xD4] = new OpCB0xD4();
		cbcodes[0xD5] = new OpCB0xD5();
		cbcodes[0xD6] = new OpCB0xD6();
		cbcodes[0xD7] = new OpCB0xD7();
		cbcodes[0xD8] = new OpCB0xD8();
		cbcodes[0xD9] = new OpCB0xD9();
		cbcodes[0xDA] = new OpCB0xDA();
		cbcodes[0xDB] = new OpCB0xDB();
		cbcodes[0xDC] = new OpCB0xDC();
		cbcodes[0xDD] = new OpCB0xDD();
		cbcodes[0xDE] = new OpCB0xDE();
		cbcodes[0xDF] = new OpCB0xDF();
		cbcodes[0xE0] = new OpCB0xE0();
		cbcodes[0xE1] = new OpCB0xE1();
		cbcodes[0xE2] = new OpCB0xE2();
		cbcodes[0xE3] = new OpCB0xE3();
		cbcodes[0xE4] = new OpCB0xE4();
		cbcodes[0xE5] = new OpCB0xE5();
		cbcodes[0xE6] = new OpCB0xE6();
		cbcodes[0xE7] = new OpCB0xE7();
		cbcodes[0xE8] = new OpCB0xE8();
		cbcodes[0xE9] = new OpCB0xE9();
		cbcodes[0xEA] = new OpCB0xEA();
		cbcodes[0xEB] = new OpCB0xEB();
		cbcodes[0xEC] = new OpCB0xEC();
		cbcodes[0xED] = new OpCB0xED();
		cbcodes[0xEE] = new OpCB0xEE();
		cbcodes[0xEF] = new OpCB0xEF();
		cbcodes[0xF0] = new OpCB0xF0();
		cbcodes[0xF1] = new OpCB0xF1();
		cbcodes[0xF2] = new OpCB0xF2();
		cbcodes[0xF3] = new OpCB0xF3();
		cbcodes[0xF4] = new OpCB0xF4();
		cbcodes[0xF5] = new OpCB0xF5();
		cbcodes[0xF6] = new OpCB0xF6();
		cbcodes[0xF7] = new OpCB0xF7();
		cbcodes[0xF8] = new OpCB0xF8();
		cbcodes[0xF9] = new OpCB0xF9();
		cbcodes[0xFA] = new OpCB0xFA();
		cbcodes[0xFB] = new OpCB0xFB();
		cbcodes[0xFC] = new OpCB0xFC();
		cbcodes[0xFD] = new OpCB0xFD();
		cbcodes[0xFE] = new OpCB0xFE();
		cbcodes[0xFF] = new OpCB0xFF();
	}

	/**
	 * interrupts
	 */
	private final void di() {
		//System.out.println("Interrupts DISabled.");
		ime = false;
	}
	private final void ei() {
		//System.out.println("Interrupts ENabled.");
		ime = true;
	}
	
	/**
	 * 8-bit acumulator arithmetic functions
	 */
	private final void add(int val) {
		int res = AF.getH() + val;
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		AF.setH(res);
	}
	private final void adc(int val) {
		int res = AF.getH() + val + (getC() ? 1 : 0);
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		AF.setH(res);
	}
	private final void sub(int val) {
		int res = AF.getH() - val;
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(true);
		AF.setH(res);
	}
	private final void sbc(int val) {
		int res = AF.getH() - val - (getC() ? 1 : 0);
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(true);
		AF.setH(res);
	}
	private final void and(int val) {
		int res = AF.getH() & val;
		setC(false);
		setZ(res);
		setN(false);
		AF.setH(res);
	}	
	private final void or(int val) {
		int res = AF.getH() | val;
		setC(false);
		setZ(res);
		setN(false);
		AF.setH(res);
	}	
	private final void xor(int val) {
		int res = AF.getH() ^ val;
		setC(false);
		setZ(res);
		setN(false);
		AF.setH(res);
	}	
	private final void cp(int val) {
		setC(AF.getH() < val);
		setZ(AF.getH() == val);
		setN(true);
	}
	/**
	 * 8-bit register arithmetic functions
	 */
	private final void incH(PgbRegister reg) {
		int res = (reg.getH() + 1);
		res &= 0xFF;
		setZ(res);
		setN(false);
		reg.setH(res);
	}
	private final void incL(PgbRegister reg) {
		int res = (reg.getL() + 1);
		res &= 0xFF;
		setZ(res);
		setN(false);
		reg.setL(res);
	}
	private final void incI(PgbRegister reg) {
		int res = reg.getI(mem) + 1;
		res &= 0xFF;
		setZ(res);
		setN(false);
		reg.setI(res, mem);
	}
	private final void decH(PgbRegister reg) {
		int res = reg.getH() - 1;
		res &= 0xFF;
		setZ(res);
		setN(true);
		reg.setH(res);
	}
	private final void decL(PgbRegister reg) {
		int res = reg.getL() - 1;
		res &= 0xFF;
		setZ(res);
		setN(true);
		reg.setL(res);
	}
	private final void decI(PgbRegister reg) {
		int res = reg.getI(mem) - 1;
		res &= 0xFF;
		setZ(res);
		setN(true);
		reg.setI(res, mem);
	}

	/**
	 * 16-bit register opcode functions
	 */
	private final void incR(PgbRegister reg) {
		reg.setR(reg.getR() + 1 & 0xFFFF);
	}
	private final void decR(PgbRegister reg) {
		reg.setR(reg.getR() - 1 & 0xFFFF);
	}
	private final void addR(PgbRegister reg0, PgbRegister reg1) {
		int res = reg0.getR() + reg1.getR();
		setC(res != (res &= 0xFFFF));
		reg0.setR(res);
	}
	
	/**
	 * stack
	 */
	private final void pop(PgbRegister reg) {
		reg.setL(mem.read(SP.data++));
		reg.setH(mem.read(SP.data++));
	}
	private final void push(PgbRegister reg) {
		// arggh. some roms push with SP == 0x0000,
		// totally ruining the elegance of this expression
		mem.write(--SP.data & 0xFFFF, reg.getH());
		SP.data &= 0xFFFF;
		mem.write(--SP.data, reg.getL());
	}
	
	/**
	 * jumping around...
	 */
	private final void jp(int address) {
		PC.setR(address);
	}
	private final void jr(int offset) {
		PC.setR(PC.getR() + (byte)offset);
	}
	private final void call(int address) {
		push(PC);
		PC.setR(address);
	}
	private final void rst(int address) {
		push(PC);
		PC.setR(address);
	}
	private final void ret() {
		pop(PC);
	}
	
	/**
	 * bitwise operators
	 */
	private final void bit(int bit, int val) {
		int res = val & (int)(Math.pow(2, bit));
		setZ(res);
		setN(false);
	}
	private final void resH(int bit, PgbRegister reg) {
		int res = reg.getH() & ~(int)(Math.pow(2, bit));
		reg.setH(res);
	}
	private final void resL(int bit, PgbRegister reg) {
		int res = reg.getL() & ~(int)(Math.pow(2, bit));
		reg.setL(res);
	}
	private final void resI(int bit, int address) {
		int res = unsign(mem.read(address)) & ~(int)(Math.pow(2, bit));
		mem.write(address, res);
	}
	private final void setH(int bit, PgbRegister reg) {
		int res = reg.getH() | (int)(Math.pow(2, bit));
		reg.setH(res);
	}
	private final void setL(int bit, PgbRegister reg) {
		int res = reg.getL() | (int)(Math.pow(2, bit));
		reg.setL(res);
	}
	private final void setI(int bit, int address) {
		int res = unsign(mem.read(address)) | (int)(Math.pow(2, bit));
		mem.write(address, res);
	}
	
	/**
	 * rotates and shifts
	 */
	private final void rrcH(PgbRegister reg) {
		int res = reg.getH();
		setC((res & 0x01) == 1);
		res = (res >> 1) | ((res & 0x01) << 7);
		setZ(res);
		setN(false);
		reg.setH(res);
	}
	private final void rrcL(PgbRegister reg) {
		int res = reg.getL();
		setC((res & 0x01) == 1);
		res = (res >> 1) | ((res & 0x01) << 7);
		setZ(res);
		setN(false);
		reg.setL(res);
	}
	private final void rrcI(int address) {
		int res = unsign(mem.read(address));
		setC((res & 0x01) == 1);
		res = (res >> 1) | ((res & 0x01) << 7);
		setZ(res);
		setN(false);
		mem.write(address, res);
	}
	private final void rrH(PgbRegister reg) {
		int res = reg.getH();
		res = (res >> 1) | (getC() ? 0x80 : 0);
		setC((reg.getH() & 0x01) == 1);
		setZ(res);
		setN(false);
		reg.setH(res);
	}
	private final void rrL(PgbRegister reg) {
		int res = reg.getL();
		res = (res >> 1) | (getC() ? 0x80 : 0);
		setC((reg.getL() & 0x01) == 1);
		setZ(res);
		setN(false);
		reg.setL(res);
	}
	private final void rrI(int address) {
		int res = unsign(mem.read(address));
		res = (res >> 1) | (getC() ? 0x80 : 0);
		setC((unsign(mem.read(address)) & 0x01) == 1);
		setZ(res);
		setN(false);
		mem.write(address, res);
	}
	private final void rlcH(PgbRegister reg) {
		int res = reg.getH();
		res = (res << 1) | (res >> 7);
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		reg.setH(res);
	}
	private final void rlcL(PgbRegister reg) {
		int res = reg.getL();
		res = (res << 1) | (res >> 7);
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		reg.setL(res);
	}
	private final void rlcI(int address) {
		int res = unsign(mem.read(address));
		res = (res << 1) | (res >> 7);
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		mem.write(address, res);
	}
	private final void rlH(PgbRegister reg) {
		int res = (reg.getH() << 1) | (getC() ? 1 : 0);
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		reg.setH(res);
	}
	private final void rlL(PgbRegister reg) {
		int res = (reg.getL() << 1) | (getC() ? 1 : 0);
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		reg.setL(res);
	}
	private final void rlI(int address) {
		int res = (unsign(mem.read(address)) << 1) | (getC() ? 1 : 0);
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		mem.write(address, res);
	}
	private final void sraH(PgbRegister reg) {
		int res = reg.getH();
		setC((res & 1) == 1);
		res = (res >> 1) | (res & 0x80);
		setZ(res);
		setN(false);
		reg.setH(res);
	}
	private final void sraL(PgbRegister reg) {
		int res = reg.getL();
		setC((res & 1) == 1);
		res = (res >> 1) | (res & 0x80);
		setZ(res);
		setN(false);
		reg.setL(res);
	}
	private final void sraI(int address) {
		int res = unsign(mem.read(address));
		setC((res & 1) == 1);
		res = (res >> 1) | (res & 0x80);
		setZ(res);
		setN(false);
		mem.write(address, res);
	}
	private final void slaH(PgbRegister reg) {
		int res = reg.getH() << 1;
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		reg.setH(res);
	}
	private final void slaL(PgbRegister reg) {
		int res = reg.getL() << 1;
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		reg.setL(res);
	}
	private final void slaI(int address) {
		int res = unsign(mem.read(address));
		res *= 2;
		setC(res != (res &= 0xFF));
		setZ(res);
		setN(false);
		mem.write(address, res);
	}
	private final void srlH(PgbRegister reg) {
		int res = reg.getH();
		setC((res & 1) == 1);
		res >>= 1;
		setZ(res);
		setN(false);
		reg.setH(res);
	}
	private final void srlL(PgbRegister reg) {
		int res = reg.getL();
		setC((res & 1) == 1);
		res >>= 1;
		setZ(res);
		setN(false);
		reg.setL(res);
	}
	private final void srlI(int address) {
		int res = unsign(mem.read(address));
		setC((res & 1) == 1);
		res >>= 1;
		setZ(res);
		setN(false);
		mem.write(address, res);
	}
	
	/**
	 * misc operations
	 */
	private final void cpl() {
		int res = ~AF.getH() & 0xFF;
		setN(true);
		AF.setH(res);
	}
	private final void daa() {
		int res = AF.getH();
		if(getN()) {
			if(res / 16 > 9) {
				res -= 0x60;
			}
			if(res % 16 > 9) {
				res -= 0x06;
			}
		} else {
			if(res / 16 > 9) {
				res += 0x60;
			}
			if(res % 16 > 9) {
				res += 0x06;
			}
		}
		setC(res != (res &= 0xFF));
		setZ(res);
		AF.setH(res);
	}
	private final void swapH(PgbRegister reg) {
		int res = ((reg.getH() << 4) & 0xF0) | (reg.getH() >> 4);
		setC(false);
		setZ(res);
		setN(false);
		reg.setH(res);
	}
	private final void swapL(PgbRegister reg) {
		int res = ((reg.getL() << 4) & 0xF0) | (reg.getL() >> 4);
		setC(false);
		setZ(res);
		setN(false);
		reg.setL(res);
	}
	private final void swapI(int address) {
		int res = ((unsign(mem.read(address)) << 4) & 0xF0) | (unsign(mem.read(address)) >> 4);
		setC(false);
		setZ(res);
		setN(false);
		mem.write(address, res);
	}
	
	/**
	 * flaggies?
	 */
	private final void setZ(boolean zval) {
		AF.lo &= ~Z_FLAG;
		if(zval) {
			AF.lo |= Z_FLAG;
		}
	}
	private final void setZ(int zval) {
		AF.lo &= ~Z_FLAG;
		if(zval == 0) {
			AF.lo |= Z_FLAG;
		}
	}
	private final void setN(boolean nval) {
		AF.lo &= ~N_FLAG;
		if(nval) {
			AF.lo |= N_FLAG;
		}
	}
	private final void setH(boolean hval) {
		AF.lo &= ~H_FLAG;
		if(hval) {
			AF.lo |= H_FLAG;
		}
	}
	private final void setC(boolean cval) {
		AF.lo &= ~C_FLAG;
		if(cval) {
			AF.lo |= C_FLAG;
		}
	}
	private final boolean getZ() {
		return (AF.getL() & Z_FLAG) == Z_FLAG;
	}
	private final boolean getN() {
		return (AF.getL() & N_FLAG) == N_FLAG;
	}
	private final boolean getH() {
		return (AF.getL() & H_FLAG) == H_FLAG;
	}
	private final boolean getC() {
		return (AF.getL() & C_FLAG) == C_FLAG;
	}
	
	/**
	 * reads
	 */
	private final int readWord() {
		return unsign(mem.read(PC.data++)) | unsign(mem.read(PC.data++)) << 8;
	}
	
	private final int readByte() {
		return unsign(mem.read(PC.data++));
	}
	
	/**
	 * status displays
	 */
	private void unsupportedCB(int cb) {
		System.out.println("unsupported CB code: " + Integer.toHexString(cb));
		//PgbSettings.paused = true;
	}
	private void unsupported(int op) {
		System.out.println(Integer.toHexString(PC.data - 1) + " unsupported opcode: " + Integer.toHexString(op));
		//PgbSettings.paused = true;;
	}
	
	/**
	 * unsign a byte
	 */
	private final int unsign(byte b) {
		return (int)(b & 0xFF);
	}
}

/*

hee hee, here's the old switch statement that took up
30% of the CPU time.  I've updated the new once since,
so this one's probably buggy.

	private int doOp(byte opcode) {
		int v;
		switch(opcode) {
		case (byte)0x00 : // NOP
			return 4;
		case (byte)0x01 : // LD   BC,nnnn
			BC.setR(readWord());
			return 20;
		case (byte)0x02 : // LD   (BC),A
			mem.write(BC.getR(), AF.getH());
			return 8;
		case (byte)0x03 : // INC  BC
			incR(BC);
			return 8;
		case (byte)0x04 : // INC  B
			incH(BC);
			return 4;
		case (byte)0x05 : // DEC  B
			decH(BC);
			return 4;
		case (byte)0x06 : // LD   B,nn
			BC.setH(readByte());
			return 8;
		case (byte)0x07 : // RLCA
			rlcH(AF);
			setZ(false);
			return 4;
		case (byte)0x08 : // LD   (nnnn),SP
			mem.write(readWord(), SP.getR());
			return 20;
		case (byte)0x09 : // ADD  HL,BC
			addR(HL, BC);
			return 12;
		case (byte)0x0A : // LD   A,(BC)
			AF.setH(BC.getI(mem));
			return 8;
		case (byte)0x0B : // DEC  BC
			decR(BC);
			return 8;
		case (byte)0x0C : // INC  C
			incL(BC);
			return 4;
		case (byte)0x0D : // DEC  C
			decL(BC);
			return 4;
		case (byte)0x0E : // LD   C,nn
			BC.setL(readByte());
			return 8;
		case (byte)0x0F : // RRCA
			rrcH(AF);
			setZ(false);
			return 4;
		case (byte)0x10 : // STOP
			readByte();
			unsupported(opcode);
			return 4;
		case (byte)0x11 : // LD   DE,nnnn
			DE.setR(readWord());
			return 12;
		case (byte)0x12 : // LD   (DE),A
			DE.setI(AF.getH(), mem);
			return 8;
		case (byte)0x13 : // INC  DE
			incR(DE);
			return 8;
		case (byte)0x14 : // INC  D
			incH(DE);
			return 4;
		case (byte)0x15 : // DEC  D
			decH(DE);
			return 4;
		case (byte)0x16 : // LD   D,nn
			DE.setH(readByte());
			return 8;
		case (byte)0x17 : // RLA
			rlH(AF);
			setZ(false);
			return 4;
		case (byte)0x18 : // JR   disp
			PC.setR(PC.getR() + (byte)readByte() + 1);
			return 12;
		case (byte)0x19 : // ADD  HL,DE
			addR(HL, DE);
			return 16;
		case (byte)0x1A : // LD   A,(DE)
			AF.setH(DE.getI(mem));
			return 8;
		case (byte)0x1B : // DEC  DE
			decR(DE);
			return 4;
		case (byte)0x1C : // INC  E
			incL(DE);
			return 4;
		case (byte)0x1D : // DEC  E
			decL(DE);
			return 4;
		case (byte)0x1E : // LD   E,nn
			DE.setL(readByte());
			return 8;
		case (byte)0x1F : // RRA
			rrH(AF);
			setZ(false);
			return 4;
		case (byte)0x20 : // JR   NZ,disp
			if(!getZ()) {
				PC.setR(PC.getR() + (byte)readByte() + 1);
				return 12;
			} else {
				readByte(); // and NOP it!
				return 8;
			}
		case (byte)0x21 : // LD   HL,nnnn
			HL.setR(readWord());
			return 12;
		case (byte)0x22 : // LDI  (HL),A 
			HL.setI(AF.getH(), mem);
			incR(HL);
			return 16;
		case (byte)0x23 : // INC  HL
			incR(HL);
			return 8;
		case (byte)0x24 : // INC  H
			incH(HL);
			return 4;
		case (byte)0x25 : // DEC  H
			decH(HL);
			return 4;
		case (byte)0x26 : // LD   H,nn
			HL.setH(readByte());
			return 8;
		case (byte)0x27 : // DAA
			if(AF.getH() % 16 > 10) {
				AF.setH((AF.getH() + 6) & 0xFF);
			}
			return 4;
		case (byte)0x28 : // JR   Z,disp
			if(getZ()) {
				PC.setR(PC.getR() + (byte)readByte() + 1);
				return 12;
			} else {
				readByte(); // and NOP it!
				return 8;
			}
		case (byte)0x29 : // ADD  HL,HL
			addR(HL, HL);
			return 12;
		case (byte)0x2A : // LDI  A,(HL)
			AF.setH(HL.getI(mem));
			incR(HL);
			return 16;
		case (byte)0x2B : // DEC HL
			decR(HL);
			return 8;
		case (byte)0x2C : // INC  L
			incL(HL);
			return 4;
		case (byte)0x2D : // DEC  L
			decL(HL);
			return 4;
		case (byte)0x2E : // LD   L,nn
			HL.setL(readByte());
			return 8;
		case (byte)0x2F : // CPL
			cpl();
			return 1;
		case (byte)0x30 : // JR   NC,disp
			if(!getC()) {
				PC.setR(PC.getR() + (byte)readByte() + 1);
				return 12;
			} else {
				readByte(); // and NOP it!
				return 8;
			}
		case (byte)0x31 : // LD   SP,nnnn
			SP.setR(readWord());
			return 12;
		case (byte)0x32 : // LDD  (HL),A
			HL.setI(AF.getH(), mem);
			decR(HL);
			return 16;
		case (byte)0x33 : // INC  SP
			incR(SP);
			return 8;
		case (byte)0x34 : // INC  (HL)
			incI(HL);
			return 12;
		case (byte)0x35 : // DEC  (HL)
			decI(HL);
			return 12;
		case (byte)0x36 : // LD   (HL),nn
			HL.setI(readByte(), mem);
			return 12;
		case (byte)0x37 : // SCF
			setC(true);
			return 4;
		case (byte)0x38 : // JR   C,disp
			if(getC()) {
				PC.setR(PC.getR() + (byte)readByte() + 1);
				return 12;
			} else {
				readByte(); // and NOP it!
				return 8;
			}
		case (byte)0x39 : // ADD  HL,SP
			HL.setR(HL.getR() + SP.getR());
			return 12;
		case (byte)0x3A : // LDD  A,(HL)
			AF.setH(HL.getI(mem));
			decR(HL);
			return 16;
		case (byte)0x3B : // DEC  SP
			decR(SP);
			return 8;
		case (byte)0x3C : // INC  A
			incH(AF);
			return 4;
		case (byte)0x3D : // DEC  A
			decH(AF);
			return 4;
		case (byte)0x3E : // LD   A,nn
			AF.setH(readByte());
			return 8;
		case (byte)0x3F : // CCF
			setC(!getC());
			return 4;
		case (byte)0x40 : // LD   B,B
			BC.setH(BC.getH());
			return 4;
		case (byte)0x41 : // LD   B,C
			BC.setH(BC.getL());
			return 4;
		case (byte)0x42 : // LD   B,D
			BC.setH(DE.getH());
			return 4;
		case (byte)0x43 : // LD   B,E
			BC.setH(DE.getL());
			return 4;
		case (byte)0x44 : // LD   B,H 
			BC.setH(HL.getH());
			return 4;
		case (byte)0x45 : // LD   B,L 
			BC.setH(HL.getL());
			return 4;
		case (byte)0x46 : // LD   B,(HL) 
			BC.setH(HL.getI(mem));
			return 8;
		case (byte)0x47 : // LD   B,A 
			BC.setH(AF.getH());
			return 4;
		case (byte)0x48 : // LD   C,B 
			BC.setL(BC.getH());
			return 4;
		case (byte)0x49 : // LD   C,C 
			BC.setL(BC.getL());
			return 4;
		case (byte)0x4A : // LD   C,D 
			BC.setL(DE.getH());
			return 4;
		case (byte)0x4B : // LD   C,E 
			BC.setL(DE.getL());
			return 4;
		case (byte)0x4C : // LD   C,H
			BC.setL(HL.getH());
			return 4;
		case (byte)0x4D : // LD   C,L
			BC.setL(HL.getL());
			return 4;
		case (byte)0x4E : // LD   C,(HL) 
			BC.setL(HL.getI(mem));
			return 8;
		case (byte)0x4F : // LD   C,A 
			BC.setL(AF.getH());
			return 4;
		case (byte)0x50 : // LD   D,B 
			DE.setH(BC.getH());
			return 4;
		case (byte)0x51 : // LD   D,C 
			DE.setH(BC.getL());
			return 4;
		case (byte)0x52 : // LD   D,D 
			DE.setH(DE.getH());
			return 4;
		case (byte)0x53 : // LD   D,E 
			DE.setH(DE.getL());
			return 4;
		case (byte)0x54 : // LD   D,H 
			DE.setH(HL.getH());
			return 4;
		case (byte)0x55 : // LD   D,L 
			DE.setH(HL.getL());
			return 4;
		case (byte)0x56 : // LD   D,(HL)
			DE.setH(HL.getI(mem));
			return 8;
		case (byte)0x57 : // LD   D,A 
			DE.setH(AF.getH());
			return 4;
		case (byte)0x58 : // LD   E,B 
			DE.setL(BC.getH());
			return 4;
		case (byte)0x59 : // LD   E,C 
			DE.setL(BC.getL());
			return 4;
		case (byte)0x5A : // LD   E,D 
			DE.setL(DE.getH());
			return 4;
		case (byte)0x5B : // LD   E,E 
			DE.setL(DE.getL());
			return 4;
		case (byte)0x5C : // LD   E,H 
			DE.setL(HL.getH());
			return 4;
		case (byte)0x5D : // LD   E,L 
			DE.setL(HL.getL());
			return 4;
		case (byte)0x5E : // LD   E,(HL)
			DE.setL(HL.getI(mem));
			return 8;
		case (byte)0x5F : // LD   E,A
			DE.setL(AF.getH());
			return 8;
		case (byte)0x60 : // LD   H,B
			HL.setH(BC.getH());
			return 4;
		case (byte)0x61 : // LD   H,C
			HL.setH(BC.getL());
			return 4;
		case (byte)0x62 : // LD   H,D
			HL.setH(DE.getH());
			return 4;
		case (byte)0x63 : // LD   H,E
			HL.setH(DE.getL());
			return 4;
		case (byte)0x64 : // LD   H,H
			HL.setH(HL.getH());
			return 4;
		case (byte)0x65 : // LD   H,L
			HL.setH(HL.getL());
			return 4;
		case (byte)0x66 : // LD   H,(HL)
			HL.setH(HL.getI(mem));
			return 8;
		case (byte)0x67 : // LD   H,A
			HL.setH(AF.getH());
			return 4;
		case (byte)0x68 : // LD   L,B
			HL.setL(BC.getH());
			return 4;
		case (byte)0x69 : // LD   L,C
			HL.setL(BC.getL());
			return 4;
		case (byte)0x6A : // LD   L,D
			HL.setL(DE.getH());
			return 4;
		case (byte)0x6B : // LD   L,E
			HL.setL(DE.getL());
			return 4;
		case (byte)0x6C : // LD   L,H
			HL.setL(HL.getH());
			return 4;
		case (byte)0x6D : // LD   L,L
			HL.setL(HL.getL());
			return 4;
		case (byte)0x6E : // LD   L,(HL)
			HL.setL(HL.getI(mem));
			return 8;
		case (byte)0x6F : // LD   L,A
			HL.setL(AF.getH());
			return 4;
		case (byte)0x70 : // LD   (HL),B
			mem.write(HL.getR(), BC.getH());
			return 8;
		case (byte)0x71 : // LD   (HL),C
			mem.write(HL.getR(), BC.getL());
			return 8;
		case (byte)0x72 : // LD   (HL),D
			mem.write(HL.getR(), DE.getH());
			return 8;
		case (byte)0x73 : // LD   (HL),E
			mem.write(HL.getR(), DE.getL());
			return 8;
		case (byte)0x74 : // LD   (HL),H
			mem.write(HL.getR(), HL.getH());
			return 8;
		case (byte)0x75 : // LD   (HL),L
			mem.write(HL.getR(), HL.getL());
			return 8;
		case (byte)0x76 : // HALT
			if(ime && mem.IE != 0) {
				halted = true;
				//System.out.println("cpu HALTed.");
			} else {
				//System.out.println("cpu not HALTed because IME=" + ime + " and IE=" + mem.IE);
			}
			return 4;
		case (byte)0x77 : // LD   (HL),A
			mem.write(HL.getR(), AF.getH());
			return 8;
		case (byte)0x78 : // LD   A,B
			AF.setH(BC.getH());
			return 4;
		case (byte)0x79 : // LD   A,C
			AF.setH(BC.getL());
			return 4;
		case (byte)0x7A : // LD   A,D
			AF.setH(DE.getH());
			return 4;
		case (byte)0x7B : // LD   A,E
			AF.setH(DE.getL());
			return 4;
		case (byte)0x7C : // LD   A,H
			AF.setH(HL.getH());
			return 4;
		case (byte)0x7D : // LD   A,L
			AF.setH(HL.getL());
			return 4;
		case (byte)0x7E : // LD   A,(HL)
			AF.setH(HL.getI(mem));
			return 8;
		case (byte)0x7f : // LD   A,A
			AF.setH(AF.getH());
			return 4;
		case (byte)0x80 : // ADD  A,B
			add(BC.getH());
			return 4;
		case (byte)0x81 : // ADD  A,C
			add(BC.getL());
			return 4;
		case (byte)0x82 : // ADD  A,D
			add(DE.getH());
			return 4;
		case (byte)0x83 : // ADD  A,E
			add(DE.getL());
			return 4;
		case (byte)0x84 : // ADD  A,H
			add(HL.getH());
			return 4;
		case (byte)0x85 : // ADD  A,L
			add(HL.getL());
			return 4;
		case (byte)0x86 : // ADD  A,(HL)
			add(HL.getI(mem));
			return 4;
		case (byte)0x87 : // ADD  A,A
			add(AF.getH());
			return 4;
		case (byte)0x88 : // ADC  A,B
			adc(BC.getH());
			return 4;
		case (byte)0x89 : // ADC  A,C
			adc(BC.getL());
			return 4;
		case (byte)0x8A : // ADC  A,D
			adc(DE.getH());
			return 4;
		case (byte)0x8B : // ADC  A,E
			adc(DE.getL());
			return 4;
		case (byte)0x8C : // ADC  A,H
			adc(HL.getH());
			return 4;
		case (byte)0x8D : // ADC  A,L
			adc(HL.getL());
			return 4;
		case (byte)0x8E : // ADC  A,(HL)
			adc(HL.getI(mem));
			return 4;
		case (byte)0x8F : // ADC  A,A
			adc(AF.getH());
			return 4;
		case (byte)0x90 : // SUB  B
			sub(BC.getH());
			return 4;
		case (byte)0x91 : // SUB  C
			sub(BC.getL());
			return 4;
		case (byte)0x92 : // SUB  D
			sub(DE.getH());
			return 4;
		case (byte)0x93 : // SUB  E
			sub(DE.getL());
			return 4;
		case (byte)0x94 : // SUB  H
			sub(HL.getH());
			return 4;
		case (byte)0x95 : // SUB  L
			sub(HL.getL());
			return 4;
		case (byte)0x96 : // SUB  (HL)
			sub(HL.getI(mem));
			return 8;
		case (byte)0x97 : // SUB  A
			sub(AF.getH());
			return 4;
		case (byte)0x98 : // SBC  A,B
			sbc(BC.getH());
			return 4;
		case (byte)0x99 : // SBC  A,C
			sbc(BC.getL());
			return 4;
		case (byte)0x9A : // SBC  A,D
			sbc(DE.getH());
			return 4;
		case (byte)0x9B : // SBC  A,E
			sbc(DE.getL());
			return 4;
		case (byte)0x9C : // SBC  A,H
			sbc(HL.getH());
			return 4;
		case (byte)0x9D : // SBC  A,L
			sbc(HL.getL());
			return 4;
		case (byte)0x9E : // SBC  A,(HL)
			sbc(HL.getI(mem));
			return 8;
		case (byte)0x9F : // SBC  A,A
			sbc(AF.getH());
			return 4;
		case (byte)0xA0 : // AND  B
			and(BC.getH());
			return 4;
		case (byte)0xA1 : // AND  C
			and(BC.getL());
			return 4;
		case (byte)0xA2 : // AND  D
			and(DE.getH());
			return 4;
		case (byte)0xA3 : // AND  E
			and(DE.getL());
			return 4;
		case (byte)0xA4 : // AND  H
			and(HL.getH());
			return 4;
		case (byte)0xA5 : // AND  L
			and(HL.getL());
			return 4;
		case (byte)0xA6 : // AND  (HL)
			and(HL.getI(mem));
			return 8;
		case (byte)0xA7 : // AND  A
			and(AF.getH());
			return 4;
		case (byte)0xA8 : // XOR  B
			xor(BC.getH());
			return 4;
		case (byte)0xA9 : // XOR  C
			xor(BC.getL());
			return 4;
		case (byte)0xAA : // XOR  D
			xor(DE.getH());
			return 4;
		case (byte)0xAB : // XOR  E
			xor(DE.getL());
			return 4;
		case (byte)0xAC : // XOR  H
			xor(HL.getH());
			return 4;
		case (byte)0xAD : // XOR  L
			xor(HL.getL());
			return 4;
		case (byte)0xAE : // XOR  (HL)
			xor(HL.getI(mem));
			return 8;
		case (byte)0xAF : // XOR  A
			xor(AF.getH());
			return 4;
		case (byte)0xB0 : // OR   B
			or(BC.getH());
			return 4;
		case (byte)0xB1 : // OR   C
			or(BC.getL());
			return 4;
		case (byte)0xB2 : // OR   D
			or(DE.getH());
			return 4;
		case (byte)0xB3 : // OR   E
			or(DE.getL());
			return 4;
		case (byte)0xB4 : // OR   H
			or(HL.getH());
			return 4;
		case (byte)0xB5 : // OR   L
			or(HL.getL());
			return 4;
		case (byte)0xB6 : // OR   (HL)
			or(HL.getI(mem));
			return 8;
		case (byte)0xB7 : // OR   A
			or(AF.getH());
			return 4;
		case (byte)0xB8 : // CP   B
			cp(BC.getH());
			return 4;
		case (byte)0xB9 : // CP   C
			cp(BC.getL());
			return 4;
		case (byte)0xBA : // CP   D
			cp(DE.getH());
			return 4;
		case (byte)0xBB : // CP   E
			cp(DE.getL());
			return 4;
		case (byte)0xBC : // CP   H
			cp(HL.getH());
			return 4;
		case (byte)0xBD : // CP   L
			cp(HL.getL());
			return 4;
		case (byte)0xBE : // CP   (HL)
			cp(HL.getI(mem));
			return 8;
		case (byte)0xBF : // CP   A
			cp(AF.getH());
			return 4;
			
		case (byte)0xC0 : // RET  NZ
			if(!getZ()) {
				pop(PC);
				return 20;
			} else {
				return 8;
			}
		case (byte)0xC1 : // POP  BC
			pop(BC);
			return 12;
		case (byte)0xC2 : // JP   NZ,nnnn
			if(!getZ()) {
				PC.setR(readWord());
				return 16;
			} else {
				readWord();
				return 12;
			}
		case (byte)0xC3 : // JP nnnn
			PC.setR(readWord());
			return 16;
		case (byte)0xC4 : // CALL NZ,nnnn
			if(!getZ()) {
				v = readWord();
				push(PC);
				PC.setR(v);
				return 24;
			} else {
				readWord();
				return 12;
			}
		case (byte)0xC5 : // PUSH BC
			push(BC);
			return 16;
		case (byte)0xC6 : // ADD  A,nn
			add(readByte());
			return 8;
		case (byte)0xC7 : // RST  00H
			push(PC);
			PC.setR(0x0000);
			return 16;
		case (byte)0xC8 : // RET Z
			if(getZ()) {
				pop(PC);
				return 20;
			} else {
				return 8;
			}
		case (byte)0xC9 : // RET
			pop(PC);
			return 16;
		case (byte)0xCA : // JP   Z,nnnn
			if(getZ()) {
				PC.setR(readWord());
				return 16;
			} else {
				readWord();
				return 12;
			}
		case (byte)0xCB : // --special--
			doCB((byte)readByte());
			return 8;
		case (byte)0xCC : // CALL Z,nnnn
			if(getZ()) {
				v = readWord();
				push(PC);
				PC.setR(v);
				return 24;
			} else {
				readWord();
				return 12;
			}
		case (byte)0xCD : // CALL nnnn
			v = readWord();
			push(PC);
			PC.setR(v);
			return 24;
		case (byte)0xCE : // ADC  A,nn
			adc(readByte());
			return 8;
		case (byte)0xCF : // RST  8H
			push(PC);
			PC.setR(0x0008);
			return 16;
		case (byte)0xD0 : // RET  NC
			if(!getC()) {
				pop(PC);
				return 20;
			} else {
				return 8;
			}
		case (byte)0xD1 : // POP  DE
			pop(DE);
			return 10;
		case (byte)0xD2 : // JP   NC,nnnn
			if(!getC()) {
				PC.setR(readWord());
				return 20;
			} else {
				readWord();
				return 8;
			}
		case (byte)0xD3 : // -
			unsupported(opcode);
			return 4;
		case (byte)0xD4 : // CALL NC,nnnn
			if(!getC()) {
				v = readWord();
				push(PC);
				PC.setR(v);
				return 24;
			} else {
				readWord();
				return 12;
			}
		case (byte)0xD5 : // PUSH DE
			push(DE);
			return 16;
		case (byte)0xD6 : // SUB  nn
			sub(readByte());
			return 8;
		case (byte)0xD7 : // RST  10H
			push(PC);
			PC.setR(0x0010);
			return 16;
		case (byte)0xD8 : // RET  C
			if(getC()) {
				pop(PC);
				return 20;
			} else {
				return 8;
			}
		case (byte)0xD9 : // RETI
			pop(PC);
			ei();
			return 16;
		case (byte)0xDA : // JP   C,nnnn
			if(getC()) {
				PC.setR(readWord());
				return 16;
			} else {
				readWord();
				return 12;
			}
		case (byte)0xDB : // -
			unsupported(opcode);
			return 4;
		case (byte)0xDC : // CALL C,nnnn
			if(getC()) {
				v = readWord();
				push(PC);
				PC.setR(v);
				return 24;
			} else {
				readWord();
				return 12;
			}
		case (byte)0xDD : // -
			unsupported(opcode);
			return 4;
		case (byte)0xDE : // SBC  A,nn
			sbc(readByte());
			return 8;
		case (byte)0xDF : // RST  18H
			push(PC);
			PC.setR(0x0018);
			return 16;
		case (byte)0xE0 : // LD   ($FF00+nn),A
			mem.write(0xFF00 + readByte(), AF.getH());
			return 12;
		case (byte)0xE1 : // POP  HL
			pop(HL);
			return 12;
		case (byte)0xE2 : // LD   ($FF00+C),A
			mem.write(0xFF00 + BC.getL(), AF.getH());
			return 8;
		case (byte)0xE3 : // -
			unsupported(opcode);
			return 4;
		case (byte)0xE4 : // -
			unsupported(opcode);
			return 4;
		case (byte)0xE5 : // PUSH HL
			push(HL);
			return 16;
		case (byte)0xE6 : // AND  nn
			and(readByte());
			return 8;
		case (byte)0xE7 : // RST  20H
			push(PC);
			PC.setR(0x0020);
			return 16;
		case (byte)0xE8 : // ADD  SP,dd
			SP.setR(SP.getR() + (byte)readByte());
			setZ(false);
			return 16;
		case (byte)0xE9 : // JP   (HL)
			PC.setR(HL.getR());
			return 4;
		case (byte)0xEA : // LD   (nnnn),A     
			mem.write(readWord(), AF.getH());
			return 16;
		case (byte)0xEB : // -
			unsupported(opcode);
			return 4;
		case (byte)0xEC : // -
			unsupported(opcode);
			return 4;
		case (byte)0xED : // -
			unsupported(opcode);
			return 4;
		case (byte)0xEE : // XOR  nn
			xor(readByte());
			return 8;
		case (byte)0xEF : // RST  28H
			push(PC);
			PC.setR(0x0028);
			return 16;
		case (byte)0xF0 : // LD   A,($FF00+nn) 
			AF.setH(mem.read(0xFF00 + readByte()));
			return 8;
		case (byte)0xF1 : // POP  AF
			pop(AF);
			return 12;
		case (byte)0xF2 : // LD   A,(C) 
			AF.setH(mem.read(0xFF00 + BC.getL()));
			return 8;
		case (byte)0xF3 : // DI
			di();
			return 4;
		case (byte)0xF4 : // -
			unsupported(opcode);
			return 4;
		case (byte)0xF5 : // PUSH AF
			push(AF);
			return 16;
		case (byte)0xF6 : // OR   nn
			or(readByte());
			return 8;
		case (byte)0xF7 : // RST  30H
			push(PC);
			PC.setR(0x0030);
			return 16;
		case (byte)0xF8 : // LD   HL,SP+dd
			HL.setR(SP.getR() + (byte)readByte());
			return 12;
		case (byte)0xF9 : // LD   SP,HL
			SP.setR(HL.getR());
			return 8;
		case (byte)0xFA : // LD   A,(nnnn)
			AF.setH(mem.read(readWord()));
			return 16;
		case (byte)0xFB : // ??? EI ???
			ei();
			return 4;
		case (byte)0xFC : // -
			unsupported(opcode);
			return 4;
		case (byte)0xFD: // EI
			ei();
			return 4;
		case (byte)0xFE : // CP   nn
			cp(readByte());
			return 8;
		case (byte)0xFF : // RST  38H
			push(PC);
			PC.setR(0x0038);
			return 16;
		default:
			System.out.println("unprogrammed:" + Integer.toHexString(opcode));
			PgbSettings.paused = true;
			return 4;
		}
	}
*/